__all__ = ('GuildJoinRequest', 'GuildJoinRequestDeleteEvent', 'GuildRequestFormResponse',)

import reprlib

from scarletio import copy_docs

from ..bases import EventBase
from ..core import GUILDS
from ..user import User, create_partial_user_from_id
from ..utils import DATETIME_FORMAT_CODE, timestamp_to_datetime

from .preinstanced import GuildJoinRequestStatus


class GuildRequestFormResponse:
    """
    Form response of a guild join request.
    
    Attributes
    ----------
    """
    def __new__(cls, data):
        """
        Creates a new guild request form response instance from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items.
        """
        
        # Nothing yet
        
        self = object.__new__(cls)
        return self
    
    
    def __repr__(self):
        """Returns the guild join request form response's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # nothing yet
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __eq__(self, other):
        """Returns whether the two join requests responses are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        # Nothing yet
        
        return True
    
    
    def __hash__(self):
        """Returns the guild join request form response's hash value."""
        return 0


class GuildJoinRequestDeleteEvent(EventBase):
    """
    Represents a guild join request delete event.
    
    Attributes
    ----------
    guild_id : `int`
        The respective guild's identifier.
    user_id : `int`
        The respective user's identifier.
    """
    __slots__ = ('guild_id', 'user_id',)
    
    def __new__(cls, data):
        """
        Creates a new  guild join request delete event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received data.
        """
        guild_id = int(data['guild_id'])
        user_id = int(data['user_id'])
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.user_id = user_id
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = [
            '<', self.__class__.__name__,
            ' guild_id=', repr(self.guild_id),
            ', user_id=', repr(self.user_id),
            '>',
        ]
        
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.user_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        return self.guild_id ^ self.user_id
    
    
    @property
    def guild(self):
        """
        Returns the respective guild of the join request. The guild must be cached.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        return GUILDS.get(self.guild_id)
    
    
    @property
    def user(self):
        """
        Returns the respective user. Might return a partial user if not cached.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)


class GuildJoinRequest(GuildJoinRequestDeleteEvent):
    """
    Guild join request received withing guild join request events.
    
    Attributes
    ----------
    guild_id : `int`
        The respective guild's identifier.
    actioned_by : `None`, ``ClientUserBase``
        The user who action the join request.
    actioned_at : `None`, `datetime`
        When the join request was actioned at.
    created_at : `datetime`
        When the join request was created.
    last_seen : `None`, `datetime`
        When the user was last seen.
    rejection_reason : `None`, `str`
        The reason why the join request was rejected.
    status : ``GuildJoinRequestStatus``
        The join request's status.
    user : ``ClientUserBase``
        The user in context.
    """
    __slots__ = (
        'actioned_by', 'actioned_at', 'created_at', 'form_responses', 'last_seen', 'rejection_reason', 'status', 'user'
    )
    
    def __new__(cls, data):
        """
        Creates a new guild join request.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Received data.
        """
        actioned_by = data.get('actioned_by_user', None)
        if (actioned_by is not None):
            actioned_by = User.from_data(actioned_by)
        
        actioned_at = data.get('actioned_at', None)
        if (actioned_at is not None):
            actioned_at = timestamp_to_datetime(actioned_at)
        
        created_at = timestamp_to_datetime(data['created_at'])
        
        form_response_datas = data['form_responses']
        if form_response_datas:
            form_responses = tuple(
                GuildRequestFormResponse(form_response_data) for form_response_data in form_response_datas
            )
        else:
            form_responses = None
        

        last_seen = data.get('last_seen', None)
        if (last_seen is not None):
            last_seen = timestamp_to_datetime(last_seen)
        
        status = GuildJoinRequestStatus.get(data['application_status'])
        
        user = User.from_data(data['user'])
        
        self = GuildJoinRequestDeleteEvent.__new__(cls, data)
        self.actioned_by = actioned_by
        self.actioned_at = actioned_at
        self.created_at = created_at
        self.form_responses = form_responses
        self.last_seen = last_seen
        self.rejection_reason = data.get('rejection_reason', None)
        self.status = status
        self.user = user
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', user=')
        repr_parts.append(repr(self.user))
        
        repr_parts.append(', created_at=')
        repr_parts.append(format(self.created_at, DATETIME_FORMAT_CODE))
        
        repr_parts.append(', status=')
        repr_parts.append(self.status.name)
        
        actioned_by = self.actioned_by
        if (actioned_by is not None):
            repr_parts.append(', actioned_by=')
            repr_parts.append(repr(actioned_by))
        
        actioned_at = self.actioned_at
        if (actioned_at is not None):
            repr_parts.append(', actioned_at=')
            repr_parts.append(format(actioned_at, DATETIME_FORMAT_CODE))
        
        last_seen = self.last_seen
        if (last_seen is not None):
            repr_parts.append(', last_seen=')
            repr_parts.append(format(last_seen, DATETIME_FORMAT_CODE))
        
        form_responses = self.form_responses
        if (form_responses is not None):
            repr_parts.append(' form_responses=[')
            
            form_response_count = len(form_responses)
            index = 0
            while True:
                form_response = form_responses[index]
                repr_parts.append(repr(form_response))
                
                index += 1
                if index == form_response_count:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        rejection_reason = self.rejection_reason
        if (rejection_reason is not None):
            repr_parts.append(', rejection_reason=')
            repr_parts.append(reprlib.repr(rejection_reason))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.user_id
        yield self.status
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.actioned_by is not other.actioned_by:
            return False
        
        if self.actioned_at != other.actioned_at:
            return False
        
        if self.created_at != other.created_at:
            return False
        
        if self.form_responses != other.form_responses:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.last_seen != other.last_seen:
            return False
        
        if self.rejection_reason != other.rejection_reason:
            return False
        
        if self.status != other.status:
            return False
        
        # user ignored in favor of using `user_id`
        
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # actioned_by
        actioned_by = self.actioned_by
        if (actioned_by is not None):
            hash_value ^= actioned_by.id
        
        # actioned_at
        actioned_at = self.actioned_at
        if (actioned_at is not None):
            hash_value ^= hash(actioned_at)
        
        # created_at
        hash_value ^= hash(self.created_at)
        
        # form_responses
        form_responses = self.form_responses
        if (form_responses is not None):
            hash_value ^= len(form_responses)
            
            for form_response in form_responses:
                hash_value ^= hash(form_response)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # last_seen
        last_seen = self.last_seen
        if (last_seen is not None):
            hash_value ^= hash(last_seen)
        
        # rejection_reason
        rejection_reason = self.rejection_reason
        if (rejection_reason is not None):
            hash_value += hash(rejection_reason)
        
        # status
        hash_value ^= hash(self.status)
        
        # user
        # Ignored in favor of using `user_id`
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value



__all__ = ()

RESPONSE_FLAG_DEFERRING = 1 << 0
RESPONSE_FLAG_DEFERRED = 1 << 1
RESPONSE_FLAG_RESPONDING = 1 << 2
RESPONSE_FLAG_RESPONDED = 1 << 3
RESPONSE_FLAG_EPHEMERAL = 1 << 4

RESPONSE_FLAG_NONE = 0
RESPONSE_FLAG_ACKNOWLEDGING = RESPONSE_FLAG_DEFERRING | RESPONSE_FLAG_RESPONDING
RESPONSE_FLAG_ACKNOWLEDGED = RESPONSE_FLAG_DEFERRED | RESPONSE_FLAG_RESPONDED
RESPONSE_FLAG_DEFERRING_OR_DEFERRED = RESPONSE_FLAG_DEFERRING | RESPONSE_FLAG_DEFERRED
RESPONSE_FLAG_RESPONDING_OR_RESPONDED = RESPONSE_FLAG_RESPONDING | RESPONSE_FLAG_RESPONDED
RESPONSE_FLAG_ACKNOWLEDGING_OR_ACKNOWLEDGED = RESPONSE_FLAG_ACKNOWLEDGING | RESPONSE_FLAG_ACKNOWLEDGED



__all__ = ('ApplicationCommandCountUpdate', 'GuildUserChunkEvent', 'VoiceServerUpdateEvent', 'WebhookUpdateEvent',)

from scarletio import copy_docs, set_docs

from ...env import CACHE_PRESENCE

from ..application_command import ApplicationCommandTargetType
from ..bases import EventBase
from ..channel import ChannelType, create_partial_channel_from_id
from ..core import GUILDS
from ..user import User


class GuildUserChunkEvent(EventBase):
    """
    Represents a processed `GUILD_MEMBERS_CHUNK` dispatch event.
    
    Attributes
    ----------
    count : `int`
        The total number of chunk responses what Discord sends for the respective gateway.
    guild_id : `int`
        The guild's identifier, what received the user chunk.
    index : `int`
        The index of the received chunk response (0 <= index < count).
    nonce : `None`, `str`
        A nonce to identify guild user chunk response.
    users : `list` of ``ClientUserBase``
        The received users.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, users, nonce, index, count = event
    ```
    """
    __slots__ = ('count', 'guild_id', 'index', 'nonce', 'users')
    
    if CACHE_PRESENCE:
        def __new__(cls, data):
            guild_id = int(data['guild_id'])
            
            users = []
            for guild_profile_data in data['members']:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id)
                users.append(user)
            
            try:
                presence_datas = data['presences']
            except KeyError:
                pass
            else:
                try:
                    guild = GUILDS[guild_id]
                except KeyError:
                    pass
                else:
                    guild._apply_presences(presence_datas)
            
            self = object.__new__(GuildUserChunkEvent)
            self.guild_id = guild_id
            self.users = users
            self.nonce = data.get('nonce', None)
            self.index = data.get('chunk_index', 0)
            self.count = data.get('chunk_count', 1)
            
            return self
    
    else:
        def __new__(cls, data):
            guild_id = int(data['guild_id'])
            
            users = []
            for guild_profile_data in data['members']:
                user = User.from_data(guild_profile_data['user'], guild_profile_data, guild_id)
                users.append(user)
            
            self = object.__new__(GuildUserChunkEvent)
            self.guild_id = guild_id
            self.users = users
            self.nonce = data.get('nonce', None)
            self.index = data.get('chunk_index', 0)
            self.count = data.get('chunk_count', 1)
            
            return self
    
    set_docs(__new__,
        """
        Creates a new guild user chunk event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Guild user chunk event data.
        """
    )
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        count = self.count
        if count != 1:
            repr_parts.append(', index=')
            repr_parts.append(repr(self.index))
            repr_parts.append('/')
            repr_parts.append(repr(count))
        
        repr_parts.append(', user count=')
        repr_parts.append(repr(len(self.users)))
        
        nonce = self.nonce
        if (nonce is not None):
            repr_parts.append(' nonce=')
            repr_parts.append(repr(nonce))
        
        repr_parts.append('>')
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 5
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.users
        yield self.nonce
        yield self.index
        yield self.count
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.count != other.count:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.index != other.index:
            return False
        
        if self.nonce != other.nonce:
            return False
        
        if self.users != other.users:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # count
        hash_value ^= self.count
        
        # guild_id
        hash_value ^= self.guild_id
        
        # index
        hash_value ^= (self.index << 12)
        
        # nonce
        nonce = self.nonce
        if (nonce is not None):
            hash_value ^= hash(nonce)
        
        # users
        users = self.users
        if users:
            hash_value ^= (len(users) << 24)
            
            for user in users:
                hash_value ^= user.id
        
        return hash_value
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)


class VoiceServerUpdateEvent(EventBase):
    """
    Represents a `VOICE_SERVER_UPDATE` event.
    
    Attributes
    ----------
    endpoint : `None`, `str`
        The voice server's host.
    guild_id : `int`
        The respective guild's identifier.
    token : `None`, `str`
        Voice connection token.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, endpoint, token = event
    ```
    """
    __slots__ = ('endpoint', 'guild_id', 'token')
    
    def __new__(cls, data):
        """
        Creates a new voice server update event from the given voice server update data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Voice server update data.
        """
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.guild_id = guild_id
        self.endpoint = data.get('endpoint', None)
        self.token = data.get('token', None)
        
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        endpoint = self.endpoint
        if (endpoint is not None):
            repr_parts.append(', endpoint=')
            repr_parts.append(repr(endpoint))
        
        token = self.token
        if (token is not None):
            repr_parts.append(', token=')
            repr_parts.append(repr(self.token))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.endpoint
        yield self.token
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.endpoint != other.endpoint:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.token != other.token:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # endpoint
        endpoint = self.endpoint
        if (endpoint is not None):
            hash_value ^= hash(endpoint)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # token
        token = self.token
        if (token is not None):
            hash_value ^= hash(token)
        
        return hash_value
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)


class WebhookUpdateEvent(EventBase):
    """
    Represents a `WEBHOOKS_UPDATE` event.
    
    Attributes
    ----------
    channel_id : `int`
        The respective channel's identifier.
    guild_id : `int`
        The respective guild's identifier.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, channel_id = event
    ```
    """
    __slots__ = ('channel_id', 'guild_id')
    
    def __new__(cls, data):
        """
        Creates a new webhook update event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Webhook update data.
        """
        channel_id = int(data['channel_id'])
        
        guild_id = data.get('guild_id', None)
        # non guild webhook check ??
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.channel_id = channel_id
        self.guild_id = guild_id
        
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', channel_id=')
        repr_parts.append(repr(self.channel_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.channel_id
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # channel_id
        hash_value ^= self.channel_id
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    @property
    def channel(self):
        """
        Returns the event's channel from cache.
        
        Returns
        -------
        channel : ``Channel``
        """
        return create_partial_channel_from_id(self.channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)



class ApplicationCommandCountUpdate(EventBase):
    """
    Represents a `GUILD_APPLICATION_COMMAND_INDEX_UPDATE` event.
    
    Attributes
    ----------
    counts : `dict` of (``ApplicationCommandTargetType``, `int`) items
        The count of the application commands by target type.
    guild_id : `int`
        The respective guild's identifier.
    
    Examples
    --------
    The event can be unpacked like:
    
    ```py
    guild_id, channel_id = event
    ```
    """
    __slots__ = ('counts', 'guild_id')
    
    def __new__(cls, data):
        """
        Creates a new application command count update event from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Application command count update data.
        """
        raw_counts = data['application_command_counts']
        
        counts = {
            ApplicationCommandTargetType.get(int(application_command_target_type_string)): count
            for application_command_target_type_string, count in raw_counts.items()
        }
        
        guild_id = data.get('guild_id', None)
        # non guild check ??
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        self.counts = counts
        self.guild_id = guild_id
        
        return self
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__,]
        
        repr_parts.append(' guild_id=')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', counts={')
        
        items = sorted(self.counts.items())
        item_count = len(items)
        if item_count:
            index = 0
            while True:
                application_command_target_type, count = items[index]
                repr_parts.append(application_command_target_type.name)
                repr_parts.append(': ')
                repr_parts.append(str(count))
                
                index += 1
                if index == item_count:
                    break
                
                repr_parts.append(', ')
                continue
        
        repr_parts.append('}')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 2
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.counts
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # counts
        if self.counts != other.counts:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # counts
        for application_command_target_type, count in self.counts.items():
            hash_value ^= application_command_target_type * count * count
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    @property
    def guild(self):
        """
        Returns the event's guild from cache.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)



__all__ = ('ChannelMetadataBase',)

from re import I as re_ignore_case, compile as re_compile, escape as re_escape

from scarletio import RichAttributeErrorBaseType, include

from ...bases import Icon, IconType, PlaceHolder, IconSlot
from ...http import urls as module_urls
from ...permission.permission import PERMISSION_NONE
from ...utils import id_to_datetime

from .constants import AUTO_ARCHIVE_DEFAULT, NAME_DEFAULT
from .flags import ChannelFlag
from .preinstanced import SortOrder, VideoQualityMode, VoiceRegion


Client = include('Client')


CHANNEL_METADATA_ICON_SLOT = IconSlot(
    'icon',
    'icon',
    module_urls.channel_group_icon_url,
    module_urls.channel_group_icon_url_as,
)


class ChannelMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for channel metadatas.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()
    
    order_group = 0
    
    def __new__(cls, keyword_parameters):
        """
        Creates a new partially channel metadata.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters to work with.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        self = cls._create_empty()
        self._set_attributes_from_keyword_parameters(keyword_parameters)
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creating channel metadatas.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data receive from Discord.
        
        Returns
        -------
        self : instance<cls>
        """
        self = object.__new__(cls)
        
        self._update_attributes(data)
        
        return self
    
    
    def __repr__(self):
        """Returns the channel metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the channel metadata's hash value"""
        return 0
    
    
    def __eq__(self, other):
        """Returns whether the two channel metadatas are the same."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Compares the channel metadata's attributes to the other one's.
        
        Parameters
        ----------
        other : `instance<type<self>>`
            The other channel metadata to compare self to.
            
            > Must have the same type as self.
        
        Returns
        -------
        equal : `bool`
            Whether both metadata has the same attributes.
        """
        return True
    
    
    def _created(self, channel_entity, client):
        """
        Called when the channel entity is initialized.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The parent channel entity.
        client : `None`, ``Client``
            The client who received the channel payload.
        """
        pass
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Serialises the channel metadata to json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether we want to include identifiers as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) Items
        """
        return {}
    

    def _update_attributes(self, data):
        """
        Updates the channel metadata with overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        """
        pass
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the channel metadata and returns it's overwritten attributes as a `dict` with a
        `attribute-name` - `old-value` relation.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Channel data received from Discord.
        
        Returns
        -------
        old_attributes : `dict` of (`str`, `Any`) items
            All item in the returned dict is optional.
            
            Might contain the following items:
            

            +---------------------------------------+-----------------------------------------------------------+
            | Keys                                  | Values                                                    |
            +=======================================+===========================================================+
            | applied_tag_ids                       | `None`, `tuple` of `int`                                  |
            +---------------------------------------+-----------------------------------------------------------+
            | archived                              | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | archived_at                           | `None`, `datetime`                                        |
            +---------------------------------------+-----------------------------------------------------------+
            | auto_archive_after                    | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | available_tags                        | `None`, `tuple` of ``ForumTag``                           |
            +---------------------------------------+-----------------------------------------------------------+
            | bitrate                               | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | default_sort_order                    | ``SortOrder``                                             |
            +---------------------------------------+-----------------------------------------------------------+
            | default_thread_auto_archive_after     | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | default_thread_reaction               | `None`, ``Emoji``                                         |
            +---------------------------------------+-----------------------------------------------------------+
            | default_thread_slowmode               | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | flags                                 | ``ChannelFlag``                                           |
            +---------------------------------------+-----------------------------------------------------------+
            | icon                                  | ``Icon``                                                  |
            +---------------------------------------+-----------------------------------------------------------+
            | invitable                             | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | metadata                              | ``ChannelMetadataBase``                                   |
            +---------------------------------------+-----------------------------------------------------------+
            | name                                  | `str`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | nsfw                                  | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | open                                  | `bool`                                                    |
            +---------------------------------------+-----------------------------------------------------------+
            | owner_id                              | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | parent_id                             | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | permission_overwrites                 | `None`, `dict` of (`int`, ``PermissionOverwrite``) items  |
            +---------------------------------------+-----------------------------------------------------------+
            | position                              | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | region                                | `None`, ``VoiceRegion``                                   |
            +---------------------------------------+-----------------------------------------------------------+
            | slowmode                              | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | topic                                 | `None`, `str`                                             |
            +---------------------------------------+-----------------------------------------------------------+
            | type                                  | ``ChannelType``                                           |
            +---------------------------------------+-----------------------------------------------------------+
            | user_limit                            | `int`                                                     |
            +---------------------------------------+-----------------------------------------------------------+
            | video_quality_mode                    | ``VideoQualityMode``                                      |
            +---------------------------------------+-----------------------------------------------------------+
        """
        return {}
    
    
    @classmethod
    def _from_partial_data(cls, data):
        """
        Creates a channel metadata from the given partial data. Called by ``Channel._from_partial_data`` when a
        new partial channel is needed to be created.
        
        Parameters
        ----------
        data : `None`, `dict` of (`str`, `Any`) items
            Partial channel data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return cls._create_empty()
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates a partial channel from the given parameters.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    
    def _delete(self, channel_entity, client):
        """
        Called when the channel is deleted.
        
        Removes the channel's references.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        client : `None`, ``Client``
            The parent client entity,
        """
        pass
    
    
    def _get_users(self, channel_entity):
        """
        The users who can see this channel.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        return []
    
    
    def _iter_users(self, channel_entity):
        """
        Iterates over the users who can see the channel.
        
        This method is an iterable generator.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        
        Yields
        ------
        user : ``ClientUserBase``
        """
        return
        yield
    
    
    def _get_clients(self, channel_entity):
        """
        Helper class to get the channel's clients.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        
        Returns
        -------
        clients : `list` of ``Client``
        """
        clients = []
        for user in self._get_users(channel_entity):
            if isinstance(user, Client):
                clients.append(user)
        
        return clients
    
    
    def _get_user(self, channel_entity, name, default):
        """
        Tries to find the a user with the given name at the channel. Returns the first matched one.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        name : `str`
            The name to search for.
        default : `Any`
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase``, `default`
        """
        if (not 1 < len(name) < 38):
            return default
        
        users = self._get_users(channel_entity)
        
        if len(name) > 6 and name[-5] == '#':
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if user.discriminator == discriminator and user.name == name_:
                        return user
        
        if len(name) > 32:
            return default
        
        for user in users:
            if user.name == name:
                return user
        
        return default
    
    
    def _get_user_like(self, channel_entity, name, default):
        """
        Searches a user, who's name or nick starts with the given string and returns the first find.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        name : `str`
            The name to search for.
        default : `Any`
            The value what is returned when no user was found. Defaults to `None`.
        
        Returns
        -------
        user : ``ClientUserBase``, `default`
        """
        name_length = len(name)
        if name_length < 1:
            return default
        
        users = self._get_users(channel_entity)
        
        if name_length > 6:
            if name_length > 37:
                return default
            
            if name[-5] == '#':
                try:
                    discriminator = int(name[-4:])
                except ValueError:
                    pass
                else:
                    stripped_name = name[:-5]
                    for user in users:
                        if user.discriminator == discriminator and user.name == stripped_name:
                            return user
        
        if name_length > 32:
            return default
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        for user in users:
            if pattern.search(user.name) is None:
                continue
            
            return user
        
        return default
    
    
    def _get_users_like(self, channel_entity, name):
        """
        Searches the users, who's name or nick starts with the given string.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        name : `str`
            The name to search for.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
        """
        result = []
        
        name_length = len(name)
        if name_length < 1:
            return result
        
        users = self._get_users(channel_entity)
        
        if name_length > 6:
            if name_length > 37:
                return result
            
            if name[-5] == '#':
                try:
                    discriminator = int(name[-4:])
                except ValueError:
                    pass
                else:
                    stripped_name = name[:-5]
                    for user in users:
                        if user.discriminator == discriminator and user.name == stripped_name:
                            result.append(user)
                            break
        
        if name_length > 32:
            return result
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        for user in users:
            if pattern.search(user.name) is None:
                continue
            
            result.append(user)
        
        return result
    
    
    def _get_processed_name(self):
        """
        Returns the channel's name.
        
        Returns
        -------
        name : `str`
        """
        return self.name
    
    
    def _get_display_name(self):
        """
        Returns channel's display name.
        
        Returns
        -------
        display_name : `str`
        """
        return self.name
    

    def _get_permissions_for(self, channel_entity, user):
        """
        Returns the permissions for the given user at the channel.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
            
        
        See Also
        --------
        ``.cached_permissions_for`` : Cached permission calculator.
        """
        return PERMISSION_NONE
    
    
    def _get_cached_permissions_for(self, channel_entity, user):
        """
        Returns the permissions for the given user at the channel. If the user's permissions are not cached, calculates
        and stores them first.
        
        Parameters
        ----------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        user : ``UserBase``
            The user to calculate it's permissions of.
        
        Returns
        -------
        permissions : ``Permission``
            The calculated permissions.
        """
        return self._get_permissions_for(channel_entity, user)
    
    
    def _get_permissions_for_roles(self, channel_entity, roles):
        """
        Returns the channel permissions of an imaginary user who would have the listed roles.
        
        Parameters
        ----------
        roles : `tuple` of ``Role``
            The roles to calculate final permissions from.
        
        Returns
        -------
        channel_entity : ``Channel``
            The channel entity owning the metadata.
        permissions : ``Permission``
            The calculated permissions.
        """
        return PERMISSION_NONE
    
    
    @classmethod
    def precreate(cls, keyword_parameters):
        """
        Precreates the channel metadata. Each channel type exhaust it's own specific attributes from
        `keyword_parameters`.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters passed to ``Channel.precreate``
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            . If an parameter's value is incorrect.
        """
        self = cls._create_empty()
        self._set_attributes_from_keyword_parameters(keyword_parameters)
        return self
    
    
    def _get_created_at(self, channel_entity):
        """
        Returns when the channel was created.
        
        Returns
        -------
        created_at : `datetime`
        """
        return id_to_datetime(channel_entity.id)
    
    
    def _invalidate_permission_cache(self):
        """
        Invalidates the cached permissions of the channel.
        
        This method is only applicable for channel types with permission cache.
        """
        pass
    
    
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        """
        Sets shared parameters used by ``.__new__` and ``._precreate˙`.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `Any`) items
            Keyword parameters to work with.
        """
        pass
    
    # Slot place holders
    
    applied_tag_ids = PlaceHolder(
        None,
        """
        Returns the tags' identifier which have been applied to the thread. Applicable for threads of a forum.
        
        Returns
        -------
        applied_tag_ids : `None`, `tuple` of `int`
        """
    )
    
    
    archived = PlaceHolder(
        False,
        """
        Returns whether the thread is archived.
        
        If the channel is not a thread, then returns `False`.
        
        Returns
        -------
        archived : `bool`
        """
    )
    
    
    archived_at = PlaceHolder(
        None,
        """
        Returns when the thread was archived.
        
        Returns `None` if the the channel is not a thread one or if it is not archived.
        
        Returns
        -------
        archived_at : `None`, `datetime`
        """
    )
    
    
    auto_archive_after = PlaceHolder(
        0,
        """
        Returns the duration in seconds to automatically archive the thread after recent activity. Can be one of:
        `3600`, `86400`, `259200`, `604800`.
        
        Returns `3600` if the channel is not a thread one.
        
        Returns
        -------
        auto_archive_after : `int`
        """
    )
    
    
    available_tags = PlaceHolder(
        None,
        """
        Returns the available tags to assign to the child-thread channels.
        
        If the channel is not a forum channel, then returns `None`.
        
        Returns
        -------
        available_tags : `None`, `tuple` of ``ForumTag``
        """
    )
    
    
    bitrate = PlaceHolder(
        0,
        """
        Returns the bitrate (in bits) of the voice channel.
        
        If the channel has no bitrate, returns `0`.
        
        Returns
        -------
        bitrate : `int`
        """
    )
    
    
    default_sort_order = PlaceHolder(
        SortOrder.latest_activity,
        """
        Returns how the posts ordered in a forum channel by default.
        
        Returns
        -------
        default_sort_order : ``SortOrder``
        """
    )
    
    
    default_thread_auto_archive_after = PlaceHolder(
        AUTO_ARCHIVE_DEFAULT,
        """
        Returns the default duration in seconds to automatically archive the channel's thread after recent activity.
        
        Returns
        -------
        default_thread_auto_archive_after : `int`
        """
    )
    
    
    default_thread_reaction = PlaceHolder(
        None,
        """
        Returns the emoji to show in the add reaction button on a thread of the forum channel.
        
        Returns
        -------
        default_thread_reaction : ``Emoji``
        """
    )
    
    
    default_thread_slowmode = PlaceHolder(
        0,
        """
        Returns the default slowmode applied to the threads of the channel.
        
        Returns
        -------
        default_thread_slowmode : `int`
        """
    )
    
    
    flags = PlaceHolder(
        ChannelFlag(0),
        """
        Returns the channel's flags.
        
        Returns empty channel flags by default.
        
        Returns
        -------
        flags : ``ChannelFlag``
        """
    )
    
    
    icon = PlaceHolder(
        Icon(IconType.none, 0),
        """
        Returns the channel's icon.
        
        Returns
        -------
        icon : ``Icon``
        """
    )
    
    
    invitable = PlaceHolder(
        True,
        """
        Whether non-moderators can invite other non-moderators to the threads. Only applicable for private threads.
        
        Returns `True` by default.
        
        Returns
        -------
        invitable : `bool`
        """
    )
    

    name = PlaceHolder(
        NAME_DEFAULT,
        """
        Returns the channel's name.
        
        Returns
        -------
        name : `str`
        """
    )
    
    
    nsfw = PlaceHolder(
        False,
        """
        Returns whether the channel is not safe for work.
        
        Defaults to `False`.
        
        Returns
        -------
        nsfw : `bool`
        """
    )
    
    
    open = PlaceHolder(
        True,
        """
        Returns whether the thread channel is open.
        
        If the channel is not a thread one, will return `True`.
        
        Returns
        -------
        open : `bool`
        """
    )
    
    
    owner_id = PlaceHolder(
        0,
        """
        Returns the channel's owner's identifier.
        
        If the channel has no owner, then returns `0`.
        
        Returns
        -------
        owner_id : `int`
        """
    )
    
    
    parent_id = PlaceHolder(
        0,
        """
        Returns the channel's parent's identifier.
        
        If the channel has no parent, or if not applicable for the specific channel type returns `0`.
        
        Returns
        -------
        parent_id : `int`
        """
    )
    
    
    permission_overwrites = PlaceHolder(
        None,
        """
        Returns the channel's permission overwrites.
        
        If the channel has no permission overwrites returns `None`.
        
        Returns
        -------
        permission_overwrites : `None`, `dict` of (`int`, ``PermissionOverwrite``) items
        """
    )
    
    
    position = PlaceHolder(
        0,
        """
        Returns the channel's position.
        
        If the channel has no position, returns `0`.
        
        Returns
        -------
        position : `int`
        """
    )
    
    
    region = PlaceHolder(
        VoiceRegion.unknown,
        """
        Returns the voice region of the channel.
        
        If the channel has no voice region, returns `None`.
        
        Returns
        -------
        region : ``VoiceRegion``
        """
    )
    
    
    slowmode = PlaceHolder(
        0,
        """
        Returns the slowmode of the channel.
        
        If the channel has no slowmode, returns `0`.
        
        Returns
        -------
        slowmode : `int`
        """
    )
    
    
    thread_users = PlaceHolder(
        None,
        """
        Returns the users inside of the thread if any.
        
        If the channel has no users, or if it is not a thread channel, will return `None`.
        
        Returns
        -------
        thread_users : `None`, `dict` of (`int`, ``ClientUserBase``) items
        """
    )
    
    
    topic = PlaceHolder(
        None,
        """
        Returns the channel's topic.
        
        If the channel has no topic, returns `None`.
        
        Returns
        -------
        topic : `None`, `str`
        """
    )
    
    
    user_limit = PlaceHolder(
        0,
        """
        Returns the maximal amount of users, who can join the voice channel
        
        If the channel has not user limit, returns `0`.
        
        Returns
        -------
        user_limit : `int`
        """
    )
    
    
    video_quality_mode = PlaceHolder(
        VideoQualityMode.none,
        """
        Returns the video quality of the voice channel.
        
        If the channel has no video quality mode, returns `VideoQualityMode.none`.
        
        Returns
        -------
        video_quality_mode : ``VideoQualityMode``
        """
    )



__all__ = ()

import warnings

from scarletio import Compound, set_docs

from ....env import API_VERSION

from ...allowed_mentions import parse_allowed_mentions
from ...bases import maybe_snowflake, maybe_snowflake_pair
from ...channel import Channel, ChannelType, create_partial_channel_from_id
from ...channel.channel.utils import CHANNEL_GUILD_THREAD_FIELD_CONVERTERS
from ...core import CHANNELS
from ...http import DiscordHTTPClient
from ...message import Message, MessageFlag
from ...payload_building import build_create_payload
from ...sticker import Sticker
from ...user import ClientUserBase, create_partial_user_from_id, thread_user_create

from ..functionality_helpers import request_channel_thread_channels
from ..request_helpers import (
    add_file_to_message_data, get_channel_and_id, get_channel_id, get_components_data, get_guild_id,
    get_channel_guild_id_and_id, get_user_and_id, get_user_id, validate_content_and_embed,
)


MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS = MessageFlag().update_by_keys(embeds_suppressed=True)


class ClientCompoundThreadEndpoints(Compound):
    
    http : DiscordHTTPClient
    id: int
    
    
    async def guild_thread_get_all_active(self, guild):
        """
        Gets all the active threads of the given guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to get it's threads of.
            
            If the guild is given as `0`, will return an empty list.
        
        Returns
        -------
        threads : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `guild` is neither ``Guild`` nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id = get_guild_id(guild)
        if not guild_id:
            return []
        
        data = await self.http.guild_thread_get_all_active(guild_id)
        
        thread_channel_datas = data['threads']
        
        thread_channels = [
            Channel.from_data(thread_channel_data, self, guild_id) for thread_channel_data in thread_channel_datas
        ]
        
        thread_user_datas = data['members']
        for thread_user_data in thread_user_datas:
            thread_channel_id = int(thread_user_data['id'])
            try:
                thread_channel = CHANNELS[thread_channel_id]
            except KeyError:
                continue
    
            user_id = int(thread_user_data['user_id'])
            user = create_partial_user_from_id(user_id)
            
            thread_user_create(thread_channel, user, thread_user_data)
        
        return thread_channels
    
    
    async def thread_create(
        self, message_or_channel, channel_template, *, type_ = ..., channel_type = None, **keyword_parameters,
    ):
        """
        Creates a new thread derived from the given message or channel.
        
        > For private thread channels the guild needs to have level 2 boost level.
        
        This method is a coroutine.
        
        Parameters
        ----------
        message_or_channel : ``Channel``, ``Message``, `int`, `tuple` (`int`, `int`)
            The channel or message to create thread from.
            
            > If given as a channel instance, will create a private thread, else a public one.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            New (thread) channel to use as a template.
        
        channel_type : `None`, ``ChannelType``, `int` = `None`, Optional (Keyword only)
            The type of the created (thread) channel.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters either to define the template, or to overwrite specific fields' values.
        
        Other Parameters
        ----------------
        applied_tag_ids : `None`, `tuple` of (`int`, ``ForumTag``), Optional (Keyword only)
             The tags' identifier which have been applied to the thread. Applicable for threads of a forum.
        
        auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags.
        
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        open_ : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        Returns
        -------
        thread_channel : ``Channel``
            The created thread channel.
        
        Raises
        ------
        TypeError
            If any parameter's type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        # Message check order
        # 1.: Message
        # 4.: None -> raise
        # 5.: `tuple` (`int`, `int`)
        # 6.: raise
        #
        # Message cannot be detected by id, only cached ones, so ignore that case.
        
        if isinstance(message_or_channel, Channel):
            if not message_or_channel.is_threadable():
                raise TypeError(
                    f'{message_or_channel!r} do not supports thread creation.'
                )
            
            message_id = None
            channel = message_or_channel
            channel_id = channel.id
        
        elif isinstance(message_or_channel, Message):
            message_id = message_or_channel.id
            channel_id = message_or_channel.channel_id
            channel = CHANNELS.get(channel_id, None)
        
        else:
            channel_id = maybe_snowflake(message_or_channel)
            if (channel_id is not None):
                message_id = None
                channel = CHANNELS.get(channel_id, None)
            
            else:
                snowflake_pair = maybe_snowflake_pair(message_or_channel)
                if snowflake_pair is None:
                    raise TypeError(
                        f'`message_or_channel` can be `{Channel.__name__}`, `{Message.__name__}`, '
                        f'`int`, `tuple` (`int`, `int`), '
                        f'got {message_or_channel.__class__.__name__}; {message_or_channel!r}.'
                    )
                
                channel_id, message_id = snowflake_pair
                channel = CHANNELS.get(channel_id)
                # Checkout type
        
        # Checkout type
        if type_ is not ...:
            warnings.warn(
                (
                    f'`type_` parameter of `{self.__class__.__name__}.thread_create` is deprecated and will be '
                    f'removed in 2023 February. Please use `channel_type` instead.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            channel_type = type_
        
        # Checkout name
        if (channel_template is not None) and isinstance(channel_template, str) and ('name' not in keyword_parameters):
            warnings.warn(
                (
                    f'`name` parameter of `{self.__class__.__name__}.thread_create` is moved to be a keyword only '
                    f'parameter and the positional usage is deprecated and will be removed in 2023 February.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            keyword_parameters['name'] = channel_template
            channel_template = None
        
        if channel_type is None:
            if message_id is None:
                channel_type = ChannelType.guild_thread_private
            else:
                channel_type = ChannelType.guild_thread_public
        
        keyword_parameters['channel_type'] = channel_type
        
        
        data = build_create_payload(channel_template, CHANNEL_GUILD_THREAD_FIELD_CONVERTERS, keyword_parameters)
        
        if message_id is None:
            coroutine = self.http.thread_create(channel_id, data)
        else:
            coroutine = self.http.thread_create_from_message(channel_id, message_id, data)
        channel_data = await coroutine
        
        if channel is None:
            guild_id = channel_data.get('guild_id', None)
            if (guild_id is not None):
                guild_id = int(guild_id)
        else:
            guild_id = channel.guild_id
        
        return Channel.from_data(channel_data, self, guild_id)
    
    
    async def forum_thread_create(
        self,
        channel_forum,
        channel_template = None,
        content = None,
        *,
        allowed_mentions = ...,
        components = None,
        embed = None,
        file = None,
        nonce = None,
        sticker = None,
        suppress_embeds = False,
        tts = False,
        **keyword_parameters,
    ):
        """
        Creates and thread at the given `channel` with the given message fields. If there is nothing to send will
        return `None` for channel and `None` for message as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        channel_forum : ``Channel``, `int`
            The forum channel's identifier where the thread will be started.
        
        channel_template : `None`, ``Channel`` = `None`, Optional
            (Thread) channel entity to use as a template.
        
        content : `None`, `str`, ``EmbedBase``, `Any` = `None`, Optional
            The message's content if given. If given as `str` or empty string, then no content will be sent, meanwhile
            if any other non `str`, ``EmbedBase`` is given, then will be casted to string.
            
            If given as ``EmbedBase``, then is sent as the message's embed.
        
        allowed_mentions : `None`,  `str`, ``UserBase``, ``Role``, `list` of (`str`, ``UserBase``, ``Role`` )
                , Optional (Keyword only)
            Which user or role can the message ping (or everyone). Check ``parse_allowed_mentions`` for details.
        
        components : `None`, ``Component``, (`tuple`, `list`) of (``Component``, (`tuple`, `list`) of
                ``Component``) = `None`, Optional (Keyword only)
            Components attached to the message.
            
            > `components` do not count towards having any content in the message.
        
        embed : ``EmbedBase``, `list` of ``EmbedBase`` = `None`, Optional (Keyword only)
            The embedded content of the message.
            
            If `embed` and `content` parameters are both given as  ``EmbedBase``, then `TypeError` is raised.
        
        file : `None`, `Any` = `None`, Optional (Keyword only)
            A file or files to send. Check ``create_file_form`` for details.
        
        nonce : `None`, `str` = `None`, Optional (Keyword only)
            Used for optimistic message sending. Will shop up at the message's data.
        
        sticker : `None`, ``Sticker``, `int`, (`list`, `set`, `tuple`) of (``Sticker``, `int`) = `None` \
                , Optional (Keyword only)
            Sticker or stickers to send within the message.
        
        suppress_embeds : `bool` = `False`, Optional (Keyword only)
            Whether the message's embeds should be suppressed initially.
        
        tts : `bool` = `False`, Optional (Keyword only)
            Whether the message is text-to-speech.
        
        **keyword_parameters : Keyword parameters
            Additional keyword parameters to create the (thread) channel with.
        
        Other Parameters
        ----------------
        applied_tag_ids : `None`, `tuple` of (`int`, ``ForumTag``), Optional (Keyword only)
             The tags' identifier which have been applied to the thread. Applicable for threads of a forum.
        
        auto_archive_after : `int`, Optional (Keyword only)
            The default duration (in seconds) for newly created threads to automatically archive the themselves.
        
        flags : `int`, ``ChannelFlag``, Optional (Keyword only)
            The channel's flags.
        
        invitable : `bool`, Optional (Keyword only)
            Whether non-moderators can invite other non-moderators to the threads.
        
        name : `str`, Optional (Keyword only)
            The channel's name.
        
        open_ : `bool`, Optional (Keyword only)
            Whether the thread channel is open.
        
        slowmode : `int`, Optional (Keyword only)
            The channel's slowmode.
        
        Returns
        -------
        thread_channel : `None`, ``Channel``
            The created thread channel. `None` if there was nothing to send.
        message : `None`, ``Message``
            Returns `None` if there is nothing to send.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        See Also
        --------
        ``.message_create`` : Sending a message to a text channel.
        ``.thread_create`` : Create thread in a text channel.
        """
        channel, channel_id = get_channel_and_id(channel_forum, Channel.is_guild_forum)
        
        content, embed = validate_content_and_embed(content, embed, False)
        
        # Sticker check order:
        # 1.: None -> None
        # 2.: Sticker -> [sticker.id]
        # 3.: int (str) -> [sticker]
        # 4.: (list, tuple) of (Sticker, int (str)) -> [sticker.id / sticker, ...] / None
        # 5.: raise
        
        if sticker is None:
            sticker_ids = None
        else:
            sticker_ids = []
            if isinstance(sticker, Sticker):
                sticker_id = sticker.id
                sticker_ids.append(sticker_id)
            else:
                sticker_id = maybe_snowflake(sticker)
                if sticker_id is None:
                    if isinstance(sticker, (list, tuple)):
                        for sticker_element in sticker:
                            if isinstance(sticker_element, Sticker):
                                sticker_id = sticker_element.id
                            else:
                                sticker_id = maybe_snowflake(sticker_element)
                                if sticker_id is None:
                                    raise TypeError(
                                        f'`sticker` can contain only `{Sticker.__name__}`, `int` elements, '
                                        f'got {sticker_element.__class__.__name__}; {sticker_element!r}; '
                                        f'sticker={sticker!r}.'
                                    )
                            
                            sticker_ids.append(sticker_id)
                        
                        if not sticker_ids:
                            sticker_ids = None
                    else:
                        raise TypeError(
                            f'`sticker` can be `None`, `{Sticker.__name__}`, `int`, '
                            f'(`list`, `tuple`) of (`{Sticker.__name__}`, `int`), got '
                            f'{sticker.__class__.__name__}; {sticker!r}.'
                        )
                else:
                    sticker_ids.append(sticker_id)
        
        components = get_components_data(components, False)
        
        if __debug__:
            
            if (nonce is not None) and (not isinstance(nonce, str)):
                raise AssertionError(
                    f'`nonce` can be `None`, `str`, got {nonce.__class__.__name__}; {nonce!r}.'
                )
            
            if not isinstance(suppress_embeds, bool):
                raise AssertionError(
                    f'`suppress_embeds` can be `bool`, got {suppress_embeds.__class__.__name__}; {suppress_embeds!r}.'
                )
        
            if not isinstance(tts, bool):
                raise AssertionError(
                    f'`tts` can be `bool`, got {tts.__class__.__name__}, {tts!r}.'
                )
        
        # Build payload
        message_data = {}
        contains_content = False
        
        if (content is not None):
            message_data['content'] = content
            contains_content = True
        
        if (embed is not None):
            message_data['embeds'] = [embed.to_data() for embed in embed]
            contains_content = True
        
        if (sticker_ids is not None):
            message_data['sticker_ids'] = sticker_ids
            contains_content = True
        
        if (components is not None):
            message_data['components'] = components
        
        if tts:
            message_data['tts'] = True
        
        if (nonce is not None):
            message_data['nonce'] = nonce
        
        if (allowed_mentions is not ...):
            message_data['allowed_mentions'] = parse_allowed_mentions(allowed_mentions)
        
        if suppress_embeds:
            message_data['flags'] = MESSAGE_FLAG_VALUE_SUPPRESS_EMBEDS
        
        message_data = add_file_to_message_data(message_data, file, contains_content, False)
        if message_data is None:
            return None, None
        
        # Checkout name
        if (channel_template is not None) and isinstance(channel_template, str) and ('name' not in keyword_parameters):
            warnings.warn(
                (
                    f'`name` parameter of `{self.__class__.__name__}.forum_thread_create` is moved to be a keyword '
                    f'only parameter and the positional usage is deprecated and will be removed in 2023 February.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            keyword_parameters['name'] = channel_template
            channel_template = None
        
        data = build_create_payload(
            channel_template, CHANNEL_GUILD_THREAD_FIELD_CONVERTERS, keyword_parameters
        )
        
        data['message'] = message_data
        
        channel_data = await self.http.thread_create(channel_id, data)
        
        if channel is None:
            guild_id = channel_data.get('guild_id', None)
            if (guild_id is not None):
                guild_id = int(guild_id)
        else:
            guild_id = channel.guild_id
        
        thread_channel = Channel.from_data(channel_data, self, guild_id)
        
        message_data = channel_data.get('message', None)
        if (message_data is None):
            message = None
        else:
            message = thread_channel._create_new_message(message_data)
        
        return thread_channel, message
    
    
    async def thread_join(self, thread_channel):
        """
        Joins the client to the given thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to join to, or it's identifier.
        
        Raises
        ------
        TypeError
            If `thread_channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        
        await self.http.thread_join(channel_id)
    
    
    async def thread_leave(self, thread_channel):
        """
        Leaves the client to the given thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to join to, or it's identifier.
        
        Raises
        ------
        TypeError
            If `thread_channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        
        await self.http.thread_leave(channel_id)
    
    
    async def thread_user_get(self, thread_channel, user):
        """
        Gets a user's thread profile inside of a thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to get the user's thread profile of.
        user : ``ClientUserBase``, `int`
            The user to get it's thread profile of.
        
        Returns
        -------
        user : ``ClientUserBase``
            The user, who's thread profile was requested.
        
        Raises
        ------
        TypeError
            - If `thread_channel`'s type is incorrect.
            - If `user`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel, channel_id = get_channel_and_id(thread_channel, Channel.is_in_group_thread)
        user, user_id = get_user_and_id(user)
        
        thread_user_data = await self.http.thread_user_get(channel_id, user_id)
        
        if user is None:
            user = create_partial_user_from_id(user_id)
        
        if channel is None:
            channel = create_partial_channel_from_id(channel_id, ChannelType.guild_thread_public, 0)
        
        thread_user_create(channel, user, thread_user_data)
        
        return user
    
    
    async def thread_user_add(self, thread_channel, user):
        """
        Adds the user to the thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to add the user to, or it's identifier.
        user : ``ClientUserBase``, `int`
            The user to add to the the thread.
        
        Raises
        ------
        TypeError
            - If `thread_channel`'s type is incorrect.
            - If `user`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        user_id = get_user_id(user)
        
        if user_id == self.id:
            coroutine = self.http.thread_join(channel_id)
        else:
            coroutine = self.http.thread_user_add(channel_id, user_id)
        await coroutine
    
    
    async def thread_user_delete(self, thread_channel, user):
        """
        Deletes the user to the thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to remove the user from, or it's identifier.
        user : ``ClientUserBase``, `int`
            The user to remove from the thread.
        
        Raises
        ------
        TypeError
            - If `thread_channel`'s type is incorrect.
            - If `user`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        channel_id = get_channel_id(thread_channel, Channel.is_in_group_thread)
        user_id = get_user_id(user)
        
        if user_id == self.id:
            coroutine = self.http.thread_leave(channel_id)
        else:
            coroutine = self.http.thread_user_delete(channel_id, user_id)
        await coroutine
    
    
    async def thread_user_get_all(self, thread_channel):
        """
        Gets all the users of a thread channel.
        
        This method is a coroutine.
        
        Parameters
        ----------
        thread_channel : ``Channel``, `int`
            The channel to get it's users of.
        
        Returns
        -------
        users : `list` of ``ClientUserBase``
            The created users.
        
        Raises
        ------
        TypeError
            If `thread_channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        thread_channel, channel_id = get_channel_and_id(thread_channel, Channel.is_in_group_thread)
        
        thread_user_datas = await self.http.thread_user_get_all(channel_id)
        
        if thread_channel is None:
            thread_channel = create_partial_channel_from_id(channel_id, ChannelType.guild_thread_public, 0)
        
        users = []
        for thread_user_data in thread_user_datas:
            user_id = int(thread_user_data['user_id'])
            user = create_partial_user_from_id(user_id)
            users.append(user)
            
            thread_user_create(thread_channel, user, thread_user_data)
        
        return users
    
    
    if API_VERSION >= 10:
        async def channel_thread_get_all_active(self, channel):
            guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
            thread_channels = await self.guild_thread_get_all_active(guild_id)
            return [thread_channel for thread_channel in thread_channels if thread_channel.parent_id == channel_id]
    
    else:
        async def channel_thread_get_all_active(self, channel):
            guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
            return await request_channel_thread_channels(
                self, guild_id, channel_id, type(self.http).channel_thread_get_chunk_active,
            )
    
    set_docs(
        channel_thread_get_all_active,
        """
        Requests all the active threads of the given channel.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        
        Notes
        -----
        When using API v10 or later, this endpoint filters from ``.guild_thread_get_all_active`` method's return.
        Consider using that instead.
        
        Active threads can also be extracted from ``Guild.threads``.
        """
    )
    
    async def channel_thread_get_all_archived_private(self, channel):
        """
        Requests all the archived private of the given channel.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
        return await request_channel_thread_channels(
            self, guild_id, channel_id, type(self.http).channel_thread_get_chunk_archived_private,
        )
    
    
    async def channel_thread_get_all_archived_public(self, channel):
        """
        Requests all the archived public threads of the given channel.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
        return await request_channel_thread_channels(
            self, guild_id, channel_id, type(self.http).channel_thread_get_chunk_archived_public,
        )
    
    
    async def channel_thread_get_all_self_archived(self, channel):
        """
        Requests all the archived private threads by the client.
        
        Parameters
        ----------
        channel : ``Channel``, `int`
            The channel to request the thread of, or it's identifier.
        
        Returns
        -------
        thread_channels : `list` of ``Channel``
        
        Raises
        ------
        TypeError
            If `channel`'s type is incorrect.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild_id, channel_id = get_channel_guild_id_and_id(channel, Channel.is_in_group_threadable)
        return await request_channel_thread_channels(
            self, guild_id, channel_id, type(self.http).channel_thread_get_chunk_self_archived,
    )



__all__ = ('ChannelMetadataGuildBase',)

from re import I as re_ignore_case, compile as re_compile, escape as re_escape

from scarletio import copy_docs, export, include

from ...core import GUILDS
from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL

from .fields import parse_name, parse_parent_id, put_name_into, put_parent_id_into, validate_name, validate_parent_id

from .base import ChannelMetadataBase


Client = include('Client')


@export
class ChannelMetadataGuildBase(ChannelMetadataBase):
    """
    Base guild channel metadata type.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('_permission_cache', 'name', 'parent_id')
    
    
    @copy_docs(ChannelMetadataBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataBase.__hash__(self)
        
        # name
        name = self.name
        if name:
            hash_value ^= hash(name)
        
        # parent_id
        hash_value ^= self.parent_id
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataGuildBase, cls).from_data(data)
        
        self._permission_cache = None
        
        return self
    

    @copy_docs(ChannelMetadataBase._delete)
    def _delete(self, channel_entity, client):
        self.permission_overwrites.clear()
        self._permission_cache = None
    
    
    @copy_docs(ChannelMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        
        # name
        if self.name != other.name:
            return False
        
        # parent_id
        if self.parent_id != other.parent_id:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # name
        put_name_into(self.name, data, defaults)
        
        # parent_id
        put_parent_id_into(self.parent_id, data, defaults)
        
        return data
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._from_partial_data)
    def _from_partial_data(cls, data):
        self = super(ChannelMetadataGuildBase, cls)._from_partial_data(data)
        
        if (data is not None):
            name = data.get('name', None)
            if (name is not None):
                self.name = name
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._update_attributes)
    def _update_attributes(self, data):
        ChannelMetadataBase._update_attributes(self, data)
        
        self._permission_cache = None
        
        # name
        self.name = parse_name(data)
        
        # parent_id
        self.parent_id = parse_parent_id(data)
        

    @copy_docs(ChannelMetadataBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = ChannelMetadataBase._difference_update_attributes(self, data)
        
        self._permission_cache = None
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # parent_id
        parent_id = parse_parent_id(data)
        if self.parent_id != parent_id:
            old_attributes['parent_id'] = self.parent_id
            self.parent_id = parent_id
        
        return old_attributes
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataGuildBase, cls)._create_empty()
        
        self._permission_cache = None
        self.parent_id = 0
        self.name = ''
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._get_users)
    def _get_users(self, channel_entity):
        return [*self._iter_users(channel_entity)]
    
    
    @copy_docs(ChannelMetadataBase._iter_users)
    def _iter_users(self, channel_entity):
        guild = channel_entity.guild
        if (guild is not None):
            for user in guild.users.values():
                if self._get_permissions_for(channel_entity, user) & PERMISSION_MASK_VIEW_CHANNEL:
                    yield user
    
    
    @copy_docs(ChannelMetadataBase._get_clients)
    def _get_clients(self, channel_entity):
        guild = channel_entity.guild
        if guild is None:
            return []
        
        return guild.clients
    
    
    @copy_docs(ChannelMetadataBase._get_user)
    def _get_user(self, channel_entity, name, default):
        name_length = len(name)
        if (name_length < 1) or (name_length > 32):
            return
        
        guild_id = channel_entity.guild_id
        if guild_id not in GUILDS:
            return
        
        users = self._get_users(channel_entity)
        
        if (name_length > 6) and (name[-5] == '#'):
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if (user.discriminator == discriminator) and (user.name == name_):
                        return user
        
        if name_length > 32:
            return default

        for user in users:
            if user.name == name:
                return user
        
        for user in users:
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            nick = guild_profile.nick
            if nick is None:
                continue
            
            if nick == name:
                return user
        
        return default
    
    
    @copy_docs(ChannelMetadataBase._get_user_like)
    def _get_user_like(self, channel_entity, name, default):
        name_length = len(name)
        if (name_length < 1) or (name_length > 37):
            return
        
        guild_id = channel_entity.guild_id
        if guild_id not in GUILDS:
            return
        
        users = self._get_users(channel_entity)
        
        if (name_length > 6) and (name[-5] == '#'):
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if (user.discriminator == discriminator) and (user.name == name_):
                        return user
        
        if (name_length > 32):
            return default
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        
        for user in users:
            if pattern.search(user.name) is not None:
                return user
            
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            nick = guild_profile.nick
            if nick is None:
                continue
            
            if pattern.search(nick) is None:
                continue
            
            return user
        
        return default


    @copy_docs(ChannelMetadataBase._get_users_like)
    def _get_users_like(self, channel_entity, name):
        result = []
        
        name_length = len(name)
        if (name_length < 1) or (name_length > 37):
            return result
        
        guild_id = channel_entity.guild_id
        if guild_id not in GUILDS:
            return
        
        users = self._get_users(channel_entity)
        
        if (name_length > 6) and (name[-5] == '#'):
            try:
                discriminator = int(name[-4:])
            except ValueError:
                pass
            else:
                name_ = name[:-5]
                for user in users:
                    if (user.discriminator == discriminator) and (user.name == name_):
                        result.append(user)
                        break
        
        if name_length > 32:
            return result
        
        pattern = re_compile(re_escape(name), re_ignore_case)
        
        for user in users:
            if pattern.search(user.name) is not None:
                result.append(user)
                continue
            
            try:
                guild_profile = user.guild_profiles[guild_id]
            except KeyError:
                continue
            
            nick = guild_profile.nick
            if nick is None:
                continue
            
            if pattern.search(nick) is None:
                continue
            
            result.append(user)
        
        return result
    
    
    @copy_docs(ChannelMetadataBase._get_cached_permissions_for)
    def _get_cached_permissions_for(self, channel_entity, user):
        if not isinstance(user, Client):
            return self._get_permissions_for(channel_entity, user)
        
        permission_cache = self._permission_cache
        if permission_cache is None:
            self._permission_cache = permission_cache = {}
        else:
            try:
                return permission_cache[user.id]
            except KeyError:
                pass
        
        permissions = self._get_permissions_for(channel_entity, user)
        permission_cache[user.id] = permissions
        return permissions
    
    
    @copy_docs(ChannelMetadataBase._set_attributes_from_keyword_parameters)
    def _set_attributes_from_keyword_parameters(self, keyword_parameters):
        ChannelMetadataBase._set_attributes_from_keyword_parameters(self, keyword_parameters)
        
        # name
        try:
            name = keyword_parameters.pop('name')
        except KeyError:
            pass
        else:
            self.name = validate_name(name)
        
        # parent_id
        try:
            parent_id = keyword_parameters.pop('parent_id')
        except KeyError:
            pass
        else:
            self.parent_id = validate_parent_id(parent_id)
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._invalidate_permission_cache)
    def _invalidate_permission_cache(self):
        self._permission_cache = None



__all__ = ('SlashCommandCategory',)

from functools import partial as partial_func

from scarletio import RichAttributeErrorBaseType, WeakReferer, copy_docs, include

from .....discord.application_command import ApplicationCommandOption, ApplicationCommandOptionType
from .....discord.application_command.constants import APPLICATION_COMMAND_OPTIONS_MAX
from .....discord.client import Client
from .....discord.events.handling_helpers import Router, _EventHandlerManager, create_event_from_class
from .....discord.interaction import InteractionEvent

from ...exceptions import (
    SlashCommandParameterConversionError, _register_exception_handler, handle_command_exception, test_exception_handler
)

from ..command_base import CommandBase
from ..command_base_application_command.constants import (
    APPLICATION_COMMAND_CATEGORY_DEEPNESS_MAX, APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND,
    APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY
)

from .helpers import _build_auto_complete_parameter_names, _register_auto_complete_function, _reset_parent_schema
from .slash_command_parameter_auto_completer import SlashCommandParameterAutoCompleter


Slasher = include('Slasher')
SlashCommand = include('SlashCommand')


class SlashCommandCategory(RichAttributeErrorBaseType):
    """
    Represents an application command's backend implementation.
    
    Attributes
    ----------
    _auto_completers : `None`, `list` of ``SlashCommandParameterAutoCompleter``
        Auto completer functions by.
    _deepness : `int`
        How nested the category is.
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _self_reference : ``WeakReferer`` to ``SlashCommandCategory``
        Back reference to the slasher application command category.
        
        Used by sub commands to access the parent entity.
    
    _sub_commands : `dict` of (`str`, ``SlashCommandFunction``) items
        The sub-commands of the category.
    _parent_reference : `None`, ``WeakReferer`` to ``SlashCommand``
        The parent slash command of the category if any.
    description : `str`
        The slash command's description.
    default : `bool`
        Whether the command is the default command in it's category.
    name : `str`
        The name of the slash sub-category.
    """
    __slots__ = (
        '__weakref__', '_auto_completers', '_deepness', '_exception_handlers', '_self_reference', '_sub_commands',
        '_parent_reference', 'description', 'default', 'name'
    )
    
    def __new__(cls, slasher_application_command, deepness):
        """
        Creates a new ``SlashCommandCategory`` with the given parameters.
        
        Parameters
        ----------
        slasher_application_command : ``SlashCommand``
            The parent slash command.
        """
        if deepness > APPLICATION_COMMAND_CATEGORY_DEEPNESS_MAX:
            raise RuntimeError('Cannot add anymore sub-category under sub-categories.')
        
        self = object.__new__(cls)
        self.name = slasher_application_command.name
        self.description = slasher_application_command.description
        self._sub_commands = {}
        self._parent_reference = None
        self.default = slasher_application_command.default
        self._auto_completers = None
        self._deepness = deepness
        self._exception_handlers = None
        self._self_reference = None
        
        return self
    
    
    async def invoke(self, client, interaction_event, options):
        """
        Calls the slash command category.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        options : `None`, `list` of ``InteractionEventChoice``
            Options bound to the category.
        """
        if (options is None) or len(options) != 1:
            return
        
        option = options[0]
        
        try:
            sub_command = self._sub_commands[option.name]
        except KeyError:
            pass
        else:
            await sub_command.invoke(client, interaction_event, option.options)
            return
        
        # Do not put this into the `except` branch.
        await handle_command_exception(
            self,
            client,
            interaction_event,
            SlashCommandParameterConversionError(
                None,
                option.name,
                'sub-command',
                list(self._sub_commands.keys()),
            )
        )
        return
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the respective auto completion function of the command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        auto_complete_option : ``ApplicationCommandAutocompleteInteraction``, \
                ``ApplicationCommandAutocompleteInteractionOption``
            The option to autocomplete.
        """
        auto_complete_option_type = auto_complete_option.type
        if (
            (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND) or
            (auto_complete_option_type is APPLICATION_COMMAND_OPTION_TYPE_SUB_COMMAND_CATEGORY)
        ):
            options = auto_complete_option.options
            if (options is not None):
                option = options[0]
                sub_commands = self._sub_commands
                if (sub_commands is not None):
                    try:
                        sub_command = sub_commands[option.name]
                    except KeyError:
                        pass
                    else:
                        await sub_command.invoke_auto_completion(client, interaction_event, option)
    
    
    def as_option(self):
        """
        Returns the slash command category as an application command option.
        
        Returns
        -------
        option : ``ApplicationCommandOption``
        """
        sub_commands = self._sub_commands
        if sub_commands:
            options = [sub_command.as_option() for sub_command in sub_commands.values()]
        else:
            options = None
        
        return ApplicationCommandOption(
            self.name,
            self.description,
            ApplicationCommandOptionType.sub_command_group,
            options = options,
            default = self.default,
        )
    
    
    def copy(self):
        """
        Copies the slash command category.
        
        Returns
        -------
        new : ``SlashCommandCategory``
        """
        new = object.__new__(type(self))
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            auto_completers = auto_completers.copy()
        new._auto_completers = auto_completers
        
        # _deepness
        new._deepness = self._deepness
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            exception_handlers = exception_handlers.copy()
        new._exception_handlers = exception_handlers
        
        # _self_reference
        new._self_reference = None
        
        # _sub_commands
        new._sub_commands = {category_name: category.copy() for category_name, category in self._sub_commands.items()}
        
        # _parent_reference
        new._parent_reference = None
        
        # default
        new.default = self.default
        
        # description
        new.description = self.description
        
        # name
        new.name = self.name
        
        return new
    
    
    def __hash__(self):
        """Returns the slash command category's hash value."""
        hash_value = 0
        
        # _auto_completers
        auto_completers = self._auto_completers
        if (auto_completers is not None):
            hash_value ^= len(auto_completers)
            
            for auto_completer in auto_completers:
                hash_value ^= hash(auto_completer)
        
        # _deepness
        # Internal field
        
        # _exception_handlers
        exception_handlers = self._exception_handlers
        if (exception_handlers is not None):
            hash_value ^= len(exception_handlers) << 4
            
            for exception_handler in exception_handlers:
                try:
                    exception_handler_hash_value = hash(exception_handler)
                except TypeError:
                    exception_handler_hash_value = object.__hash__(exception_handler)
                hash_value ^= exception_handler_hash_value
        
        # _self_reference
        # Internal field
        
        # _sub_commands
        sub_commands = self._sub_commands
        if (sub_commands is not None):
            hash_value ^= len(sub_commands) << 8
            
            for sub_command in sub_commands:
                hash_value ^= hash(sub_command)
        
        # _parent_reference
        # Internal field
        
        # default
        hash_value ^= self.default << 12
        
        # description
        description = self.description
        hash_value ^= hash(description)
        
        # name
        name = self.name
        if name != description:
            hash_value ^= hash(name)
        
        return hash_value
    

    def __format__(self, code):
        """Formats the command in a format string."""
        if not code:
            return str(self)
        
        if code == 'm':
            return self.mention
        
        raise ValueError(
            f'Unknown format code {code!r} for {self.__class__.__name__}; {self!r}. '
            f'Available format codes: {""!r}, {"m"!r}.'
        )
    
    
    @property
    def interactions(self):
        """
        Enables you to add sub-commands under the sub-category.
        
        Returns
        -------
        handler : ``_EventHandlerManager``
        """
        return _EventHandlerManager(self)
    
    
    def create_event(self, func, *args, **kwargs):
        """
        Adds a sub-command under the slash category.
        
        Parameters
        ----------
        func : `async-callable`
            The function used as the command when using the respective slash command.
        *args : Positional Parameters
            Positional parameters to pass to ``SlashCommand``'s constructor.
        **kwargs : Keyword parameters
            Keyword parameters to pass to the ``SlashCommand``'s constructor.
        
        Returns
        -------
        self : ``SlashCommandCategory``
        
        Raises
        ------
        TypeError
            If Any parameter's type is incorrect.
        ValueError
            If Any parameter's value is incorrect.
        RuntimeError
            - The ``SlashCommand`` reached the maximal amount of children.
            - Cannot add anymore sub-category under sub-categories.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        if isinstance(func, Router):
            func = func[0]
        
        if isinstance(func, SlashCommand):
            self._add_application_command(func)
            return self
        
        command = SlashCommand(func, *args, **kwargs)
        if isinstance(command, Router):
            command = command[0]
        
        return self._add_application_command(command)
    
    
    def create_event_from_class(self, klass):
        """
        Breaks down the given class to it's class attributes and tries to add it as a sub-command.
        
        Parameters
        ----------
        klass : `type`
            The class, from what's attributes the command will be created.
        
        Returns
        -------
        self : ``SlashCommandCategory``
         
        Raises
        ------
        TypeError
            If Any attribute's type is incorrect.
        ValueError
            If Any attribute's value is incorrect.
        RuntimeError
            - The ``SlashCommand`` reached the maximal amount of children.
            - Cannot add anymore sub-category under sub-categories.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        command = create_event_from_class(
            SlashCommand, klass, SlashCommand.COMMAND_PARAMETER_NAMES, SlashCommand.COMMAND_NAME_NAME,
            SlashCommand.COMMAND_COMMAND_NAME
        )
        
        if isinstance(command, Router):
            command = command[0]
        
        return self._add_application_command(command)
    
    
    def _add_application_command(self, command):
        """
        Adds a sub-command or sub-category to the slash command.
        
        Parameters
        ----------
        command : ``SlashCommand``
            The slash command to add.
        
        Returns
        -------
        as_sub_command : ``SlashCommandFunction``, ``SlashCommandCategory``
            The command as sub-command.
        
        Raises
        ------
        RuntimeError
            - The ``SlashCommand`` reached the maximal amount of children.
            - Cannot add anymore sub-category under sub-categories.
            - If the command to add is a default sub-command meanwhile the category already has one.
        """
        sub_commands = self._sub_commands
        if len(sub_commands) == APPLICATION_COMMAND_OPTIONS_MAX and (command.name not in sub_commands):
            raise RuntimeError(
                f'The {self.__class__.__name__} reached the maximal amount of children '
                f'({APPLICATION_COMMAND_OPTIONS_MAX}).'
            )
        
        as_sub_command = command.as_sub_command(self._deepness + 1)
        
        if command.default:
            for sub_command in sub_commands.values():
                if sub_command.default:
                    raise RuntimeError(
                        f'{self!r} already has default command.'
                    )
        
        as_sub_command._parent_reference = self._get_self_reference()
        sub_commands[command.name] = as_sub_command
        
        _reset_parent_schema(self)
        
        # Resolve auto completers recursively
        parent = self
        while True:
            auto_completers = parent._auto_completers
            if (auto_completers is not None):
                for auto_completer in auto_completers:
                    as_sub_command._try_resolve_auto_completer(auto_completer)
            
            if isinstance(parent, Slasher):
                break
            
            parent_reference = parent._parent_reference
            if (parent_reference is None):
                break
            
            parent = parent_reference()
            if (parent is None):
                break
        
        return as_sub_command
    
    
    def __eq__(self, other):
        """Returns whether the two slash commands categories are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # _auto_completers
        if self._auto_completers != other._auto_completers:
            return False
        
        # _deepness
        # Internal Field
        
        # _exception_handlers
        if self._exception_handlers != other._exception_handlers:
            return False
        
        # _self_reference
        # Internal Field
        
        # _sub_commands
        if self._sub_commands != other._sub_commands:
            return False
        
        # _parent_reference
        # Internal Field
        
        # default
        if self.default != other.default:
            return False
        
        # description
        if self.description != other.description:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        return True
    
    
    def autocomplete(self, parameter_name, *parameter_names, function=None):
        """
        Registers an auto completer function to the application command.
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        *parameter_names : `str`
            Additional parameter names to autocomplete
        function : `None`, `callable` = `None`, Optional (Keyword only)
            The function to register as auto completer.
        
        Returns
        -------
        function / wrapper : `async-callable`, `functools.partial`
            The registered function if given or a wrapper to register the function with.
        """
        parameter_names = _build_auto_complete_parameter_names(parameter_name, parameter_names)
        
        if (function is None):
            return partial_func(_register_auto_complete_function, self, parameter_names)
            
        return self._add_autocomplete_function(parameter_names, function)
    
    
    def _add_autocomplete_function(self, parameter_names, function):
        """
        Registers an autocomplete function.
        
        Parameters
        ----------
        parameter_names : `list` of `str`
            The parameters' names.
        function : `async-callable`
            The function to register as auto completer.
        
        Returns
        -------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The registered auto completer
        
        Raises
        ------
        RuntimeError
            - If the application command function has no parameter named, like `parameter_name`.
            - If the parameter cannot be auto completed.
        TypeError
            - If `function` is not an asynchronous.
        """
        if isinstance(function, SlashCommandParameterAutoCompleter):
            function = function._command
        
        auto_completer = SlashCommandParameterAutoCompleter(
            function,
            parameter_names,
            self._deepness,
            self,
        )
        
        auto_completers = self._auto_completers
        if (auto_completers is None):
            auto_completers = []
            self._auto_completers = auto_completers
        
        auto_completers.append(auto_completer)
        
        resolved = 0
        sub_commands = self._sub_commands
        for sub_command in sub_commands.values():
            resolved += sub_command._try_resolve_auto_completer(auto_completer)
        
        if resolved:
            _reset_parent_schema(self)
        
        return auto_completer
    
    
    def _try_resolve_auto_completer(self, auto_completer):
        """
        Tries to register auto completer to the slasher application command function.
        
        Parameters
        ----------
        auto_completer : ``SlashCommandParameterAutoCompleter``
            The auto completer.
        
        Returns
        -------
        resolved : `int`
            The amount of parameters resolved.
        """
        resolved = 0
        for sub_command in self._sub_commands.values():
             resolved += sub_command._try_resolve_auto_completer(auto_completer)
        
        return resolved
    
    
    def error(self, exception_handler=None, *, first=False):
        """
        Registers an exception handler to the ``SlashCommandCategory``.
        
        Parameters
        ----------
        exception_handler : `None`, `CoroutineFunction` = `None`, Optional
            Exception handler to register.
        first : `bool` = `False`, Optional (Keyword Only)
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler / wrapper : `CoroutineFunction` / `functools.partial`
            If `exception_handler` is not given, returns a wrapper.
        """
        if exception_handler is None:
            return partial_func(_register_exception_handler, first)
        
        return self._register_exception_handler(exception_handler, first)
    
    
    def _register_exception_handler(self, exception_handler, first):
        """
        Registers an exception handler to the ``SlashCommandCategory``.
        
        Parameters
        ----------
        exception_handler : `CoroutineFunction`
            Exception handler to register.
        first : `bool`
            Whether the exception handler should run first.
        
        Returns
        -------
        exception_handler : `CoroutineFunction`
        """
        test_exception_handler(exception_handler)
        
        exception_handlers = self._exception_handlers
        if exception_handlers is None:
            self._exception_handlers = exception_handlers = []
        
        if first:
            exception_handlers.insert(0, exception_handler)
        else:
            exception_handlers.append(exception_handler)
        
        return exception_handler
    
    
    def _get_self_reference(self):
        """
        Gets a weak reference to the ``SlashCommandCategory``.
        
        Returns
        -------
        self_reference : ``WeakReferer`` to ``SlashCommandCategory``
        """
        self_reference = self._self_reference
        if self_reference is None:
            self_reference = WeakReferer(self)
            self._self_reference = self_reference
        
        return self_reference
    
    
    # ---- Mention ----
    
    @property
    @copy_docs(CommandBase.mention)
    def mention(self):
        return self._mention_recursive()
    
    
    def _mention_recursive(self, *sub_command_names):
        """
        Returns the application command category's mention.
        
        Called by ``.mention`` to include the sub-commands' names.
        
        Parameters
        ----------
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        parent_reference = self._parent_reference
        if parent_reference is None:
            parent = None
        else:
            parent = parent_reference()
        
        if parent is None:
            return ''
        
        return parent._mention_recursive(self.name, *sub_command_names)
    
    
    @copy_docs(CommandBase.mention_at)
    def mention_at(self, guild):
        return self._mention_at_recursive(guild)
    
    
    def _mention_at_recursive(self, guild, *sub_command_names):
        """
        Returns the application command category's mention.
        
        Called by ``.mention`` to include the sub-commands' names.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to mention the command at.
        
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        parent_reference = self._parent_reference
        if parent_reference is None:
            parent = None
        else:
            parent = parent_reference()
        
        if parent is None:
            return ''
        
        return parent._mention_at_recursive(guild, self.name, *sub_command_names)



__all__ = ('remove_default_plugin_variables',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.remove_default_variables)
def remove_default_plugin_variables(*names):
    return PLUGIN_LOADER.remove_default_variables(*names)



__all__ = ('EMOJI_ALL_RP',)

from re import compile as re_compile, escape as re_escape, U as re_unicode

from scarletio import call

from .unicodes import UNICODES
from ..utils import EMOJI_RP

# We import builtins, so this file is loaded after that. Bad loading order can happen when running tests.
from . import unicodes


def trie_node_sort_key(node):
    """
    Sort key for trie nodes.
    
    Parameters
    ----------
    node : ``TrieNode``
        The node to get sort key of.
    
    Returns
    -------
    sort_key : `str`
    """
    character = node.character
    if character is None:
        character = ''
    
    return character


def escaped_string_iterator(string):
    """
    Regex escapes each character in a string and iterates over them.
    
    Parameters
    ----------
    string : `str`
        The string to iterate over.
    
    Yields
    ------
    escaped : `str`
    """
    for character in string:
        yield re_escape(character)


class TrieNode:
    """
    Represents a pattern node when building a pattern.
    
    Attributes
    ----------
    character : `None`, `str`
        The represented character.
    end : `bool`
        Whether self is end of a node.
    nodes : `None`, `dict` of (`str`, ``PatternNode``) items
        The nodes inheriting from self.
    """
    __slots__ = ('character', 'end', 'nodes')
    
    def __new__(cls, character):
        """
        Creates a new trie node.
        
        Parameters
        ----------
        character : `None`, `str`
            The represented character,
        """
        self = object.__new__(cls)
        self.character = character
        self.end = False
        self.nodes = None
        return self
    
    def has_length_1(self):
        """
        Returns whether length of self is 1.
        
        Returns
        -------
        has_length_1 : `bool`
        """
        if self.character is None:
            return False
        
        if self.nodes is not None:
            return False
        
        return True
    
    
    def extend_with_raw_string(self, string):
        """
        Extends self with the given string.
        
        Parameters
        ----------
        string : `str`
            The string to add to self.
        """
        return self.extend_with_escaped_iterator(escaped_string_iterator(string))
    
    
    def extend_with_escaped_iterator(self, iterator):
        """
        Extends the pattern with the given string iterator.
        
        Parameters
        ----------
        iterator : `GeneratorType`
            The iterator to walk through.
        """
        character = next(iterator, None)
        if character is None:
            self.end = True
            return
        
        nodes = self.nodes
        if nodes is None:
            nodes = {}
            self.nodes = nodes
        
        try:
            node = nodes[character]
        except KeyError:
            node = type(self)(character)
            nodes[character] = node
        
        node.extend_with_escaped_iterator(iterator)
    
    
    def build_into(self, into):
        """
        Builds self into the given list.
        
        Parameters
        ----------
        into : `list` of `str`
            The list to extend with the pattern's parents.
        
        Returns
        -------
        into : `list` of `str`
        """
        character = self.character
        if (character is not None):
            into.append(character)
        
        nodes = self.nodes
        if (nodes is not None):
            nodes = sorted(nodes.values(), key = trie_node_sort_key)
            
            node_count = len(nodes)
            if node_count == 1:
                node = nodes[0]
                
                if node.has_length_1():
                    into = node.build_into(into)
                    if self.end:
                        into.append('?')
                else:
                    if self.end:
                        into.append('(?:')
                    
                    into = node.build_into(into)
                    
                    if self.end:
                        into.append(')?')
            
            else:
                into.append('(?:')
                
                node_index = 0
                
                while True:
                    node = nodes[node_index]
                    node_index += 1
                    
                    into = node.build_into(into)
                    
                    if node_index == node_count:
                        break
                    
                    into.append('|')
                    continue
                
                into.append(')')
                
                if self.end:
                    into.append('?')
        
        return into


EMOJI_ALL_RP = EMOJI_RP

@call
def build_all_emoji_pattern():
    """
    Builds an emoji matching pattern which matches not only the custom, but the custom emojis as well.
    
    Returns
    -------
    pattern : `re.Pattern`
    """
    global EMOJI_ALL_RP
    
    into = []
    into.append('(?:(')
    
    pattern_core = TrieNode(None)
    for unicode in UNICODES:
        pattern_core.extend_with_raw_string(unicode.value)
    pattern_core.build_into(into)
    pattern_core = None
    
    into.append(')|(?<!\\\\)\:(')
    
    pattern_core = TrieNode(None)
    for unicode in UNICODES:
        pattern_core.extend_with_raw_string(unicode.name)
        for alias in unicode.iter_aliases():
            pattern_core.extend_with_raw_string(alias)
    
    pattern_core.build_into(into)
    pattern_core = None
    
    into.append(')\:|')
    into.append(EMOJI_RP.pattern)
    into.append(')')
    
    EMOJI_ALL_RP = re_compile(''.join(into), re_unicode)



__all__ = ('ActivityFieldBase',)

from scarletio import RichAttributeErrorBaseType


class ActivityFieldBase(RichAttributeErrorBaseType):
    """
    Base class for activity fields.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new activity field.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new activity field from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Activity field data.
        """
        return object.__new__(cls)
    
    
    def to_data(self):
        """
        Serializes the activity field.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        return {}
    
    
    def __repr__(self):
        """Returns the activity field's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __eq__(self, other):
        """Returns whether the two activity fields are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        return True
    
    
    def __hash__(self):
        """Returns the activity field's hash value."""
        return 0
    
    
    def __bool__(self):
        """Returns whether the activity field has any non-default attribute set."""
        return False



__all__ = ('BaseSnapshotType',)

from scarletio import RichAttributeErrorBaseType, WeakReferer


SNAPSHOT_TYPES = set()


class BaseSnapshotType(RichAttributeErrorBaseType):
    """
    Base class for snapshots.
    
    Attributes
    ----------
    _client_reference : `WeakReferer` to `Client`
        Weakreference to the owner client instance.
    _is_difference : `bool`
        Whether the snapshot is a difference.
    """
    __slots__ = ('_client_reference', '_is_difference')
    
    def __new__(cls, client):
        """
        Takes a snapshot.
        
        Parameters
        ----------
        client : ``Client``
            The client to take snapshot of.
        """
        client_reference = WeakReferer(client)
        
        self = object.__new__(cls)
        self._client_reference = client_reference
        self._is_difference = False
        return self
    
    
    def __repr__(self):
        """Returns the snapshot's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        client = self.client
        if (client is not None):
            repr_parts.append(' of ')
            repr_parts.append(repr(client))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __sub__(self, other):
        """Subtracts other from self."""
        if other is None:
            return self
        
        if type(self) is not type(other):
            return NotImplemented
        
        return self._extract(other)
    
    
    def _extract(self, other):
        """
        Extracts `other` from `self` returning a new snapshot, but this time a difference.
        
        Parameters
        ----------
        other : ``BaseSnapshotType``
            Other snapshot to extract from self.
        
        Returns
        -------
        difference : ``BaseSnapshotType``
        
        """
        if (self._is_difference != other._is_difference):
            raise RuntimeError(
                f'Only same kind of snapshot can be extracted, got self={self!r}; other={other!r}.'
            )
            
        new = object.__new__(type(self))
        new._client_reference = self._client_reference
        new._is_difference = True
        return new
    
    
    def __init_subclass__(cls):
        """
        Registers the subclasses as snapshot types.
        """
        SNAPSHOT_TYPES.add(cls)
    
    
    @property
    def client(self):
        """
        Returns the snapshot's client.
        """
        return self._client_reference()
    
    
    def is_revertible(self):
        """
        Returns whether self can be reverted.
        
        Returns
        -------
        is_revertible : `bool`
        """
        if not self._is_difference:
            return False
        
        if self._client_reference() is None:
            return False
        
        return True
    
    
    def revert(self):
        """
        Reverts the snapshot.
        """
        pass
    
    
    def __bool__(self):
        """Returns whether the snapshot contains any changes."""
        return False



__all__ = ()

from ..helpers import render_main_call_into

from .render_constants import (
    BOX_BOTTOM, BOX_LEFT, BOX_LEFT_BOT, BOX_LEFT_TOP, BOX_RIGHT, BOX_RIGHT_BOT, BOX_RIGHT_TOP, BOX_TITLE_ERROR,
    BOX_TITLE_SUB_COMMANDS, BOX_TOP
)


def render_usage_line_into(into, name_trace_iterator):
    """
    Renders the command usage line into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list to render the usage into.
    name_trace_iterator : `None`, `GeneratorType`
        Name trace iterator for looking up the command's access trace.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append('Usage: ')
    render_main_call_into(into)
    
    if (name_trace_iterator is not None):
        for name in name_trace_iterator:
            into.append(' ')
            into.append(name)
    
    return into


def render_sub_command_box_into(into, sub_command_names, line_length):
    """
    Renders the sub-command box into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list to render the usage into.
    sub_command_names : `str`
        The sub command names to render.
    line_length : `str`
        The expected length of every line.
    
    Returns
    -------
    into : `list` of `str`
    """
    into = render_box_start_into(into, line_length, BOX_TITLE_SUB_COMMANDS)
    
    # Render line n
    for sub_command_name in sub_command_names:
        into.append(BOX_LEFT)
        into.append(sub_command_name)
        into = render_box_line_adjustment_into(into, len(sub_command_name), line_length)
        into.append(BOX_RIGHT)
        into.append('\n')
    
    # Render line -1
    into = render_box_end_into(into, line_length)

    return into


def render_box_start_into(into, line_length, title):
    """
    Renders box start into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to render to.
    line_length : `int`
        The expected length of the line.
    title : `str`
        Title of the box.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(BOX_LEFT_TOP)
    into.append(title)
    
    line_adjust = line_length - len(title)
    if line_adjust > 0:
        into.append(BOX_TOP * line_adjust)
    
    into.append(BOX_RIGHT_TOP)
    into.append('\n')
    
    return into


def render_box_line_adjustment_into(into, local_line_length, global_line_length):
    """
    Renders box line adjustment into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to render to.
    local_line_length : `int`
        The expected length of the line locally.
    global_line_length : `int`
        The global line length,
    
    Returns
    -------
    into : `list` of `str`
    """
    line_adjust = global_line_length - local_line_length
    if line_adjust > 0:
        into.append(' ' * line_adjust)
    
    return into


def render_box_end_into(into, line_length):
    """
    Renders box end into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to render to.
    line_length : `int`
        The expected length of the line.
    
    Returns
    -------
    into : `list` of `str`
    """
    into.append(BOX_LEFT_BOT)
    into.append(BOX_BOTTOM * line_length)
    into.append(BOX_RIGHT_BOT)
    into.append('\n')
    
    return into


def render_error_box_into_single_line(into, line):
    """
    Renders an error message box into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to render to.
    line : `str`
        The message to show.
    
    Returns
    -------
    into : `list` of `str`
    """
    return render_error_box_into_multi_line(into, [line])


def render_error_box_into_multi_line(into, lines):
    """
    Renders an error message box of multiple lines into the given list.
    
    Parameters
    ----------
    into : `list` of `str`
        The list of strings to render to.
    message : `list` of `str`
        The message lines to show.
    
    Returns
    -------
    into : `list` of `str`
    """
    line_length = max(len(line) for line in lines)
    line_length = max(line_length, len(BOX_TITLE_ERROR))
    
    into = render_box_start_into(into, line_length, BOX_TITLE_ERROR)
    
    for line in lines:
        into.append(BOX_LEFT)
        into.append(line)
        into = render_box_line_adjustment_into(into, len(line), line_length)
        into.append(BOX_RIGHT)
        into.append('\n')
    
    into = render_box_end_into(into, line_length)
    return into



__all__ = ('InteractionComponent',)

import reprlib, warnings

from scarletio import RichAttributeErrorBaseType, export

from ...component import ComponentType

from .fields import (
    parse_custom_id, parse_components, parse_type, parse_value, put_custom_id_into,
    put_components_into, put_type_into, put_value_into, validate_custom_id, validate_components, validate_type,
    validate_value
)


@export
class InteractionComponent(RichAttributeErrorBaseType):
    """
    Integration component representing a parameter or a sub-command-group.
    
    Attributes
    ----------
    components : `None`, `tuple` of ``InteractionComponent``
        Nested components.
    
    custom_id : `None`, `str`
        The `.custom_id` of the represented component.
    
    type : ``ComponentType``
        The represented component's type.
    
    value : `None`, `str`
        The components value defined by teh user.
        
        > Mutually exclusive with the `components` field.
    """
    __slots__ = ('components', 'custom_id', 'type', 'value')
    
    
    def __new__(cls, **keyword_parameters):
        """
        Creates a new interaction component from the given keyword parameters.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields and how should be set.
        
        Other Parameters
        ----------------
        components : `None`, `tuple` of ``InteractionComponent``, Optional (Keyword only)
            Nested components.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            The `.custom_id` of the represented component.
        
        type_ : ``ComponentType``, Optional (Keyword only)
            The represented component's type.
        
        value : `None`, `str`, Optional (Keyword only)
            The components value defined by teh user.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        self = cls._create_empty()
        self._populate_from_keyword_parameters(keyword_parameters)
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an interaction component with it's default values set.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.custom_id = None
        self.components = None
        self.type = ComponentType.none
        self.value = None
        return self
    
    
    def copy(self):
        """
        Copies the interaction component.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        components = self.components
        if (components is not None):
            components = tuple(component.copy() for component in components)
        new.components = components
        new.custom_id = self.custom_id
        new.type = self.type
        new.value = self.value
        return new
    
    
    def copy_with(self, **keyword_parameters):
        """
        Copies the interaction component with replacing the defined fields.
        
        Parameters
        ----------
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields and how should be set.
        
        Other Parameters
        ----------------
        components : `None`, `tuple` of ``InteractionComponent``, Optional (Keyword only)
            Nested components.
        
        custom_id : `str`, Optional (Keyword only)
            The `.custom_id` of the represented component.
        
        type_ : ``ComponentType``, Optional (Keyword only)
            The represented component's type.
        
        value : `None`, `str`, Optional (Keyword only)
            The components value defined by teh user.
        
        Returns
        -------
        new : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        self = self.copy()
        self._populate_from_keyword_parameters(keyword_parameters)
        return self
    
    
    def _populate_from_keyword_parameters(self, keyword_parameters):
        """
        Sets the interaction component's attributes from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `object`) items
            A dictionary of keyword parameters defining which fields and how should be set.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        if not keyword_parameters:
            return
        
        # components
        try:
            components = keyword_parameters.pop('components')
        except KeyError:
            pass
        else:
            self.components = validate_components(components)
        
        # custom_id
        try:
            custom_id = keyword_parameters.pop('custom_id')
        except KeyError:
            pass
        else:
            self.custom_id = validate_custom_id(custom_id)
        
        # type
        try:
            type_ = keyword_parameters.pop('type_')
        except KeyError:
            pass
        else:
            self.type = validate_type(type_)
        
        # value
        try:
            value = keyword_parameters.pop('value')
        except KeyError:
            pass
        else:
            self.value = validate_value(value)
        
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new interaction component from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`)
            Interaction component data.
        """
        self = object.__new__(cls)
        self.components = parse_components(data)
        self.custom_id = parse_custom_id(data)
        self.type = parse_type(data)
        self.value = parse_value(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the interaction component into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `Any`) items
        """
        data = {}
        put_components_into(self.components, data, defaults)
        put_custom_id_into(self.custom_id, data, defaults)
        put_type_into(self.type, data, defaults)
        put_value_into(self.value, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the interaction component's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type
        
        # type
        type_ = self.type
        if type_ is not ComponentType.none:
            repr_parts.append(' type = ')
            repr_parts.append(type_.name)
            repr_parts.append(' (')
            repr_parts.append(repr(type_.value))
            repr_parts.append(')')
            
            field_added = True
        
        else:
            field_added = False
        
        # System fields : custom_id
        
        # custom_id
        
        custom_id = self.custom_id
        if custom_id is not None:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' custom_id = ')
            repr_parts.append(reprlib.repr(custom_id))
        
        # Extra descriptive fields : components | value
        # components
        components = self.components
        if (components is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' components = [')
            
            index = 0
            limit = len(components)
            
            while True:
                option = components[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # value
        value = self.value
        if (value is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' value = ')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the interaction component's hash value."""
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= (len(components) << 8)
            
            for component in components:
                hash_value ^= hash(component)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two interaction component's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # components
        if self.components != other.components:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    @property
    def options(self):
        """
        `.options` is deprecated and will be removed in 2023 February. Please use `.components` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.options` is deprecated and will be removed in 2023 February. '
                f'Please use `.components` instead.'  
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.components
    
    
    def iter_components(self):
        """
        Iterates over the sub-components of the interaction component.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``InteractionComponent``
        """
        components = self.components
        if (components is not None):
            yield from components


    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        custom_id = self.custom_id
        if (custom_id is not None):
            yield custom_id, self.value
        
        for component in self.iter_components():
            yield from component.iter_custom_ids_and_values()



__all__ = ('create_partial_role_from_id', 'cr_p_role_object', 'parse_role', 'parse_role_mention')

import warnings
from functools import partial as partial_func

from scarletio import export

from ..core import ROLES
from ..utils import ID_RP, ROLE_MENTION_RP

from .fields import (
    put_color_into, put_mentionable_into, put_name_into, put_permissions_into, put_position_into, put_separated_into,
    put_unicode_emoji_into, validate_color, validate_mentionable, validate_name, validate_permissions,
    validate_position, validate_separated, validate_unicode_emoji
)
from .role import Role


ROLE_FIELD_CONVERTERS = {
    'color': (validate_color, put_color_into),
    'icon': (Role.icon.validate_data_icon, partial_func(Role.icon.put_into, as_data = True)),
    'mentionable': (validate_mentionable, put_mentionable_into),
    'name': (validate_name, put_name_into),
    'permissions': (validate_permissions, put_permissions_into),
    'position': (validate_position, put_position_into),
    'separated': (validate_separated, put_separated_into),
    'unicode_emoji': (validate_unicode_emoji, put_unicode_emoji_into),
}


@export
def create_partial_role_from_id(role_id, guild_id = 0):
    """
    Creates a partial role from the given `role_id`. If the role already exists returns that instead.
    
    Parameters
    ----------
    role_id : `int`
        The unique identifier number of the role.
    guild_id : `int` = `0`, Optional
        The role's guild's identifier.
    
    Returns
    -------
    role : ``Role``
    """
    try:
        return ROLES[role_id]
    except KeyError:
        pass
    
    role = Role._create_empty(role_id, guild_id)
    ROLES[role_id] = role
    
    return role


def cr_p_role_object(name, **keyword_parameters):
    """
    Deprecated, please use `Role(..).to_data(...)` instead.
    
    Will be removed in 2023 February.
    """
    warnings.warn(
        (
            f'`cr_p_role_object` is deprecated and will be removed in 2023 February. '
            f'Please use `Role(..).to_data(...)` instead.'
        ),
        FutureWarning,
        stacklevel = 2,
    )
    return Role(name = name, **keyword_parameters).to_data()


def parse_role_mention(text):
    """
    If the text is a role mention, returns the respective role if found.
    
    Parameters
    ----------
    text : `str`
       The text to parse the role out from.
    
    Returns
    -------
    role : `None`, ``Role``
        The found role if any.
    """
    parsed = ROLE_MENTION_RP.fullmatch(text)
    if parsed is None:
        return
    
    role_id = int(parsed.group(1))
    return ROLES.get(role_id, None)


def parse_role(text, message = None):
    """
    Tries to parse a role out from the given text.
    
    Parameters
    ----------
    text : `str`
        The text to parse the role out.
    
    message : `None`, ``Message`` = `None`, Optional
        Context for name based parsing.
    
    Returns
    -------
    role : `None`, ``Role``
        The found role if any.
    """
    parsed = ID_RP.fullmatch(text)
    if (parsed is not None):
        role_id = int(parsed.group(1))
        try:
            role = ROLES[role_id]
        except KeyError:
            pass
        else:
            return role
    
    role = parse_role_mention(text)
    if (role is not None):
        return role
    
    if (message is not None):
        guild = message.guild
        if (guild is not None):
            role = guild.get_role_like(text)
            if (role is not None):
                return role
    
    return None



__all__ = ('ChannelMetadataGuildCategory',)

from scarletio import copy_docs

from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildCategory(ChannelMetadataGuildMainBase):
    """
    Guild category channel metadata.
    
    Attributes
    ----------
    _permission_cache : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    parent_id : `int`
        The channel's parent's identifier.
    name : `str`
        The channel's name.
    permission_overwrites : `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    
    Class Attributes
    ----------------
    order_group: `int` = `4`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()
    
    order_group = 4
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.upper()



from .action_metadata import *
from .trigger_metadata import *

from .action import *
from .constants import *
from .event_types import *
from .preinstanced import *
from .rule import *

__all__ = (
    *action_metadata.__all__,
    *trigger_metadata.__all__,
    
    *action.__all__,
    *constants.__all__,
    *event_types.__all__,
    *preinstanced.__all__,
    *rule.__all__,
)



__all__ = ()

import reprlib, warnings

from scarletio import Compound
from scarletio.web_common import Formdata

from ...bases import maybe_snowflake_pair
from ...core import  GUILDS, STICKERS
from ...emoji import Emoji
from ...http import DiscordHTTPClient, VALID_STICKER_IMAGE_MEDIA_TYPES
from ...sticker import Sticker, StickerPack
from ...utils import MEDIA_TYPE_TO_EXTENSION, get_image_media_type

from ..functionality_helpers import ForceUpdateCache
from ..request_helpers import get_guild_and_id, get_guild_id, get_sticker_and_id, get_sticker_pack_and_id


STICKER_PACK_CACHE = ForceUpdateCache()


class ClientCompoundStickerEndpoints(Compound):
    
    http : DiscordHTTPClient
    
    
    async def sticker_get(self, sticker, *, force_update=False):
        """
        Gets an sticker by it's id. If the sticker is already loaded updates it.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The sticker, who will be requested.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker : ``Sticker``
        
        Raises
        ------
        TypeError
            If `sticker` was not given neither as ``Sticker``, neither as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        TypeError
            If `sticker` was not given as ``Sticker``, nor as `int`.
        """
        sticker, sticker_id = get_sticker_and_id(sticker)
        if (sticker is not None) and (not sticker.partial) and (not force_update):
            return sticker
        
        data = await self.http.sticker_get(sticker_id)
        if (sticker is None):
            sticker = Sticker(data)
        else:
            sticker._update_from_partial(data)
        
        return sticker
    
    
    async def sticker_pack_get(self, sticker_pack, *, force_update=False):
        """
        Gets the sticker packs. If the sticker-packs are already loaded, updates them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker_pack : ``StickerPack``, `int`
            The sticker pack' identifier.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker-pack should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker_packs : `list` of ``StickerPack``
        
        Raises
        ------
        TypeError
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        sticker_pack, sticker_pack_id = get_sticker_pack_and_id(sticker_pack)
        
        if (sticker_pack is None) or force_update:
            sticker_pack_data = await self.http.sticker_pack_get(sticker_pack_id)
            sticker_pack = StickerPack._create_and_update(sticker_pack_data)
        
        return sticker_pack
    
    
    async def sticker_pack_get_all(self, *, force_update=False):
        """
        Gets the sticker packs. If the sticker-packs are already loaded, updates them.
        
        This method is a coroutine.
        
        Parameters
        ----------
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker-packs should be requested even if it supposed to be up to date.
        
        Returns
        -------
        sticker_packs : `list` of ``StickerPack``
        
        Raises
        ------
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if force_update or (not STICKER_PACK_CACHE.synced):
            data = await self.http.sticker_pack_get_all()
            sticker_pack_datas = data['sticker_packs'] # Discord pls.
            if force_update:
                sticker_packs = [StickerPack._create_and_update(sticker_pack_data) for sticker_pack_data in
                    sticker_pack_datas]
            else:
                sticker_packs = [StickerPack(sticker_pack_data) for sticker_pack_data in sticker_pack_datas]
            
            STICKER_PACK_CACHE.set(sticker_packs)
        
        else:
            sticker_packs = STICKER_PACK_CACHE.value
        
        return sticker_packs
    
    
    async def sticker_guild_get(self, sticker, *deprecated_parameters, force_update=False):
        """
        Gets the specified sticker from the respective guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `tuple`, (`int`, `int`)
            The sticker to get.
        force_update : `bool` = `False`, Optional (Keyword only)
            Whether the sticker should be requested even if it supposed to be up to date.
        
        Raises
        ------
        TypeError
            If `sticker` is not ``Sticker``, `tuple` (`int`, `int`).
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        if deprecated_parameters:
            if len(deprecated_parameters) > 1:
                raise TypeError(
                    f'`{self.__class__.__name__}.sticker_guild_get` accepts up to `2` positional parameters, got '
                    f'{len(deprecated_parameters) + 1}.'
                )
            
            warnings.warn(
                (
                    f'2nd parameter of `{self.__class__.__name__}.sticker_guild_get` is deprecated and will be '
                    f'removed in 2022 Jun. Please pass just an `{Sticker.__name__}` or a pair of snowflake.'
                ),
                FutureWarning,
                stacklevel = 2,
            )
            
            guild, sticker, = sticker, *deprecated_parameters
            sticker, sticker_id = get_sticker_and_id(sticker)
            guild_id = get_guild_id(guild)
            
        else:
            if isinstance(sticker, Sticker):
                guild_id = sticker.guild_id
                sticker_id = sticker.id
            else:
                snowflake_pair = maybe_snowflake_pair(sticker)
                if snowflake_pair is None:
                    raise TypeError(
                        f'`sticker` can be `{Sticker.__name__}`, `tuple` (`int`, `int`), '
                        f'got {sticker.__class__.__name__}; {sticker!r}.'
                    )
                
                guild_id, sticker_id = snowflake_pair
                sticker = STICKERS.get(sticker_id, None)
        
        if (sticker is not None) and (not sticker.partial) and (not force_update):
            return sticker
        
        data = await self.http.sticker_guild_get(guild_id, sticker_id)
        if (sticker is None):
            sticker = Sticker(data)
        else:
            sticker._update_from_partial(data)
        
        return sticker
    
    
    async def sticker_guild_create(self, guild, name, image, emoji_representation, description = None, *, reason = None):
        """
        Creates a sticker in the guild.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to create the sticker in.
        name : `str`
            The sticker's name. It's length can be in range [2:32]
        emoji_representation : ``Emoji``, `str`
            The emoji representation of the sticker. Used as a tag for the sticker.
        image : `bytes-like`
            The sticker's image in bytes.
        description : `None`, `str` = `None`, Optional
            The sticker's representation. It's length can be in range [0:100]
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Returns
        -------
        sticker : ``Sticker``
            The created sticker.
        
        Raises
        ------
        TypeError
            - If `emoji_representation` is neither `str` nor ``Emoji``.
            - If `guild` is neither ``Guild`` nor `int`.
        ValueError
            If `image`s media type is neither `image/png` nor `application/json`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - If `name` is not `str`.
            - If `name`'s length is out of range [2:32].
            - If `description` is neither `None`, `str`.
            - If `description`'s length is out of range [0:100].
            - If `emoji_representation` is a custom ``Emoji``.
        """
        guild_id = get_guild_id(guild)
        
        if __debug__:
            if not isinstance(name, str):
                raise AssertionError(
                    f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                )
        
            name_length = len(name)
            if (name_length < 2) or (name_length > 32):
                raise AssertionError(
                    f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                )
        
        if __debug__:
            if (description is not None):
                if (not isinstance(description, str)):
                    raise AssertionError(
                        f'`description` can be `str`, got {description.__class__.__name__}; {description!r}.'
                    )
                
                description_length = len(description)
                if (description_length > 100):
                    raise AssertionError(
                        f'`description` length can be in range [0:100], got {description_length!r}; {description!r}.'
                    )
        
        if (description is not None) and (not description):
            description = None
        
        if isinstance(emoji_representation, str):
            tag = emoji_representation
        elif isinstance(emoji_representation, Emoji):
            if __debug__:
                if emoji_representation.is_custom_emoji():
                    raise AssertionError(
                        f'Only unicode (builtin) emojis can be used as tags, got {emoji_representation!r}.'
                    )
            
            tag = emoji_representation.name
        else:
            raise TypeError(
                f'`emoji_representation` can be `str`, `{Emoji.__name__}`, got '
                f'{emoji_representation.__class__.__name__}; {emoji_representation!r}.'
            )
        
        if __debug__:
            if not isinstance(image, (bytes, bytearray, memoryview)):
                raise TypeError(
                    f'`image` can be `None`, `bytes-like`, got {image.__class__.__name__}; {reprlib.repr(image)}.'
                )
        
        media_type = get_image_media_type(image)
        if media_type not in VALID_STICKER_IMAGE_MEDIA_TYPES:
            raise ValueError(
                f'Invalid `image` type: {media_type}, got {reprlib.repr(image)}.'
            )
        
        extension = MEDIA_TYPE_TO_EXTENSION[media_type]
        
        form_data = Formdata()
        form_data.add_field('name', name)
        
        if (description is not None):
            form_data.add_field('description', description)
        else:
            # If no description is given Discord drops back an unrelated error
            form_data.add_field('description', '')
        
        form_data.add_field('tags', tag)
        
        form_data.add_field('file', image, filename=f'file.{extension}', content_type=media_type)
        
        
        sticker_data = await self.http.sticker_guild_create(guild_id, form_data, reason)
        
        return Sticker(sticker_data)
    
    
    async def sticker_guild_edit(self, sticker, *, name=..., emoji_representation=..., description = ..., reason = None):
        """
        Edits the given guild bound sticker,
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The respective sticker.
        name : `str`, Optional (keyword only)
            New name of the sticker. It's length can be in range [2:32].
        emoji_representation : ``Emoji``, `str`, Optional (keyword Only)
            The new emoji representation of the sticker. Used as a tag for the sticker.
        description : `None`, `str`, Optional (Keyword only)
            New description for the sticker. It's length can be in range [0:100].
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            - If `sticker` is neither ``Sticker``, nor `int`.
            - If `emoji_representation` is neither `str` nor ``Emoji``.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            - Non guild bound sticker cannot be edited.
            - If `name` is not `str`.
            - If `name`'s length is out of range [2:32].
            - If `description` is neither `None`, `str`.
            - If `description`'s length is out of range [0:100].
            - If `emoji_representation` is a custom ``Emoji``.
        """
        sticker = await self.sticker_get(sticker)
        
        if __debug__:
            if not sticker.guild_id:
                raise AssertionError(
                    f'Non guild bound sticker cannot be edited, got {sticker!r}.'
                )
        
        if (name is ...):
            name = sticker.name
        else:
            if __debug__:
                if not isinstance(name, str):
                    raise AssertionError(
                        f'`name` can be `str`, got {name.__class__.__name__}; {name!r}.'
                    )
            
                name_length = len(name)
                if (name_length < 2) or (name_length > 32):
                    raise AssertionError(
                        f'`name` length can be in range [2:32], got {name_length!r}; {name!r}.'
                    )
        
        if (emoji_representation is ...):
            tag = ', '.join(sticker.tags)
        else:
            if isinstance(emoji_representation, str):
                tag = emoji_representation
            elif isinstance(emoji_representation, Emoji):
                if __debug__:
                    if emoji_representation.is_custom_emoji():
                        raise AssertionError(
                            f'Only unicode (builtin) emojis can be used as tags, got {emoji_representation!r}.'
                        )
                
                tag = emoji_representation.name
            else:
                raise TypeError(
                    f'`emoji_representation` can be `str`, `{Emoji.__name__}`, got '
                    f'{emoji_representation.__class__.__name__}; {emoji_representation!r}.'
                )
        
        if (description is ...):
            description = sticker.description
        else:
            if __debug__:
                if (description is not None):
                    if (not isinstance(description, str)):
                        raise AssertionError(
                            f'`description` can be `None`, `str`, got {description.__class__.__name__}; '
                            f'{description!r}.'
                        )
            
                    description_length = len(description)
                    if (description_length > 100):
                        raise AssertionError(
                            f'`description` length can be in range [0:100], got {description_length!r}; '
                            f'{description!r}.'
                        )
            
            if (description is None):
                description = ''
        
        data = {
            'name': name,
            'tags': tag,
            'description': description,
        }
        
        await self.http.sticker_guild_edit(sticker.guild_id, sticker.id, data, reason)
    
    
    async def sticker_guild_delete(self, sticker, *, reason = None):
        """
        Deletes the sticker.
        
        This method is a coroutine.
        
        Parameters
        ----------
        sticker : ``Sticker``, `int`
            The sticker to delete.
        reason : `None`, `str` = `None`, Optional (Keyword only)
            Will show up at the respective guild's audit logs.
        
        Raises
        ------
        TypeError
            If `sticker` is neither ``Sticker``, nor `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        AssertionError
            Non guild bound sticker cannot be edited.
        """
        sticker = await self.sticker_get(sticker)
        
        if __debug__:
            if not sticker.guild_id:
                raise AssertionError(
                    f'Non guild bound sticker cannot be edited, got {sticker!r}.'
                )
        
        await self.http.sticker_guild_delete(sticker.guild_id, sticker.id, reason)
    
    
    
    async def sticker_guild_get_all(self, guild):
        """
        Syncs the given guild's stickers with the wrapper.
        
        This method is a coroutine.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild, what's stickers will be synced.
        
        Returns
        -------
        stickers : `list` of ``Sticker``
            The guild's stickers.
        
        Raises
        ------
        TypeError
            If `guild` was not given neither as ``Guild`` nor as `int`.
        ConnectionError
            No internet connection.
        DiscordException
            If any exception was received from the Discord API.
        """
        guild, guild_id = get_guild_and_id(guild)
        
        sticker_datas = await self.http.sticker_guild_get_all(guild_id)
        
        if guild is None:
            # Do not create a partial guild, because it would have been garbage collected after leaving the function
            # anyways
            guild = GUILDS.get(guild_id, None)
        
        if guild is None:
            stickers = [Sticker(sticker_data) for sticker_data in sticker_datas]
        else:
            guild._sync_stickers(sticker_datas)
            stickers = [*guild.stickers.values()]
        
        return stickers
    
    
    async def guild_sync_stickers(self, guild):
        """
        Deprecated and will be removed in 2022 Jun. Please use ``.sticker_guild_get_all`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.guild_sync_stickers` is deprecated and will be '
                f'removed in 2022 Jun. Please use `.sticker_guild_get_all` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return await self.sticker_guild_get_all(guild)



__all__ = ('ComponentType',)

import warnings

from scarletio import class_property

from ...bases import Preinstance as P, PreinstancedBase

from ..component_metadata import (
    ComponentMetadataBase, ComponentMetadataButton, ComponentMetadataChannelSelect, ComponentMetadataMentionableSelect,
    ComponentMetadataRoleSelect, ComponentMetadataRow, ComponentMetadataStringSelect, ComponentMetadataTextInput,
    ComponentMetadataUserSelect
)


class ComponentType(PreinstancedBase):
    """
    Represents a component's type.
    
    Attributes
    ----------
    name : `str`
        The name of the component type.
    value : `int`
        The identifier value the component type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ComponentType``) items
        Stores the predefined ``ComponentType``-s. These can be accessed with their `value` as key.
    VALUE_TYPE : `type` = `int`
        The component type's type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the component types.
    
    Every predefined component type can be accessed as class attribute as well:
    
    +-----------------------+-----------------------+-------+
    | Class attribute name  | Name                  | Value |
    +=======================+=======================+=======+
    | none                  | none                  | 0     |
    +-----------------------+-----------------------+-------+
    | row                   | row                   | 1     |
    +-----------------------+-----------------------+-------+
    | button                | button                | 2     |
    +-----------------------+-----------------------+-------+
    | string_select         | string select         | 3     |
    +-----------------------+-----------------------+-------+
    | text_input            | text input            | 4     |
    +-----------------------+-----------------------+-------+
    | user_select           | user select           | 5     |
    +-----------------------+-----------------------+-------+
    | role_select           | role select           | 6     |
    +-----------------------+-----------------------+-------+
    | mentionable_select    | mentionable select    | 7     |
    +-----------------------+-----------------------+-------+
    | channel_select        | channel select        | 8     |
    +-----------------------+-----------------------+-------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ('metadata_type', )
    

    @classmethod
    def _from_value(cls, value):
        """
        Creates a new component type with the given value.
        
        Parameters
        ----------
        value : `int`
            The channel type's identifier value.
        
        Returns
        -------
        self : ``ChannelType``
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.metadata_type = ComponentMetadataBase
        
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates a new component type and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the channel type.
        name : `str`
            The default name of the channel type.
        metadata_type : `None`, `type<ComponentMetadataBase>`
            The component type's respective metadata type.
        """
        self.value = value
        self.name = name
        self.metadata_type = metadata_type
        
        self.INSTANCES[value] = self
    
    
    none = P(0, 'none', ComponentMetadataBase)
    row = P(1, 'row', ComponentMetadataRow)
    button = P(2, 'button', ComponentMetadataButton)
    string_select = P(3, 'string select', ComponentMetadataStringSelect)
    text_input = P(4, 'text input', ComponentMetadataTextInput)
    user_select = P(5, 'user select', ComponentMetadataUserSelect)
    role_select = P(6, 'role select', ComponentMetadataRoleSelect)
    mentionable_select = P(7, 'mentionable select', ComponentMetadataMentionableSelect)
    channel_select = P(8, 'channel select', ComponentMetadataChannelSelect)
    
    
    @class_property
    def select(cls):
        """
        `.select` is deprecated and will be removed in 2023 January. Please use `.string_select` instead.
        """
        warnings.warn(
            (
                f'`{cls.__name__}.select` is deprecated and will be removed in 2023 January. '
                f'Please use `.string_select` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return cls.string_select



__all__ = ()

from scarletio import Task

from ...discord.core import KOKORO

from .constants import (
    LAVALINK_KEY_EVENT_PLAYER_WEBSOCKET_CLOSED, LAVALINK_KEY_EVENT_TRACK_END, LAVALINK_KEY_EVENT_TRACK_EXCEPTION,
    LAVALINK_KEY_EVENT_TRACK_START, LAVALINK_KEY_EVENT_TRACK_STUCK, LAVALINK_KEY_GUILD_ID
)
from .event_types import (
    PlayerWebsocketClosedEvent, TrackEndEvent, TrackExceptionEvent, TrackStartEvent, TrackStuckEvent
)


def parse_track_end(client, data):
    guild_id = int(data[LAVALINK_KEY_GUILD_ID])
    
    try:
        player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    event = TrackEndEvent(player, data)
    
    Task(client.solarlink._events.track_end(client, event), KOKORO)
    

def parse_track_exception(client, data):
    guild_id = int(data[LAVALINK_KEY_GUILD_ID])
    
    try:
        player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    event = TrackExceptionEvent(player, data)
    
    Task(client.solarlink._events.track_exception(client, event), KOKORO)


def parse_track_start(client, data):
    guild_id = int(data[LAVALINK_KEY_GUILD_ID])
    
    try:
        player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    event = TrackStartEvent(player, data)
    Task(client.solarlink._events.track_start(client, event), KOKORO)



def parse_track_stuck(client, data):
    guild_id = int(data[LAVALINK_KEY_GUILD_ID])
    
    try:
        player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    event = TrackStuckEvent(player, data)
    Task(client.solarlink._events.track_stuck(client, event), KOKORO)


def parse_player_websocket_closed(client, data):
    guild_id = int(data[LAVALINK_KEY_GUILD_ID])
    
    try:
        player = client.solarlink.players[guild_id]
    except KeyError:
        return
    
    event = PlayerWebsocketClosedEvent(player, data)
    Task(client.solarlink._events.player_websocket_closed(client, event), KOKORO)


PARSERS = {
    LAVALINK_KEY_EVENT_TRACK_END: parse_track_end,
    LAVALINK_KEY_EVENT_TRACK_EXCEPTION: parse_track_exception,
    LAVALINK_KEY_EVENT_TRACK_START: parse_track_start,
    LAVALINK_KEY_EVENT_TRACK_STUCK: parse_track_stuck,
    LAVALINK_KEY_EVENT_PLAYER_WEBSOCKET_CLOSED: parse_player_websocket_closed,
}

del parse_track_end
del parse_track_exception
del parse_track_start
del parse_track_stuck
del parse_player_websocket_closed



__all__ = ('load_all_plugin',)

from scarletio import copy_docs

from ..plugin_loader import PLUGIN_LOADER


@copy_docs(PLUGIN_LOADER.load)
def load_all_plugin(*, blocking=True):
    return PLUGIN_LOADER.load_all(blocking=blocking)
